# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

This modules provides classes and functions allowing to maintain *operations*.
An operation (:py:class:`Operation`) comprises a Python callable and some
additional meta-information (:py:class:`OpMetaInfo`) that allows for automatic
input validation, input value conversion, monitoring, and inter-connection of
multiple operations.

Operations are registered in operation registries (:py:class:`OpRegistry`),
the default operation registry is accessible via the global, read-only
``OP_REGISTRY`` variable.
"""

import sys
from collections import OrderedDict
from typing import Union, Callable, Optional, Dict

import xarray as xr

from .types import ValidationError
from ..util.opmetainf import OpMetaInfo
from ..util.monitor import Monitor
from ..util.undefined import UNDEFINED
from ..util.safe import safe_eval
from ..util.process import run_subprocess
from ..util.process import ProcessOutputMonitor
from ..util.tmpfile import del_temp_file
from ..util.tmpfile import new_temp_file
from ..util.misc import object_to_qualified_name
from ..util.misc import to_list
from ..version import __version__

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

_MONITOR = OpMetaInfo.MONITOR_INPUT_NAME
_RETURN = OpMetaInfo.RETURN_OUTPUT_NAME


class Operation:
    """
    An Operation comprises a wrapped callable (e.g. function, constructor,
    lambda form) and additional meta-information about the wrapped operation
    itself and its inputs and outputs.

    :param wrapped_op: some callable object that will be wrapped.
    :param op_meta_info: operation meta information.
    """

    def __init__(self, wrapped_op: Callable, op_meta_info=None):
        if callable is None:
            raise ValueError('wrapped_op must be given')
        if not callable(wrapped_op):
            raise ValueError('wrapped_op must be callable')

        if op_meta_info is None:
            # Try unwrapping wrapped_op
            try:
                # noinspection PyUnresolvedReferences
                op_meta_info = wrapped_op.op_meta_info
                try:
                    # noinspection PyUnresolvedReferences
                    wrapped_op = wrapped_op.wrapped_op
                except AttributeError:
                    pass
            except AttributeError:
                pass

        self._wrapped_op = wrapped_op
        self._op_meta_info = \
            op_meta_info or OpMetaInfo.introspect_operation(wrapped_op)
        for attr_name in [
            '__module__', '__name__', '__qualname__', '__doc__', '__file__'
        ]:
            try:
                setattr(self, attr_name, getattr(wrapped_op, attr_name))
            except AttributeError:
                pass

    @property
    def op_meta_info(self) -> OpMetaInfo:
        """
        :return: Meta-information about the operation,
            see :py:class:`esa_climate_toolbox.core.op.OpMetaInfo`.
        """
        return self._op_meta_info

    @property
    def wrapped_op(self) -> Callable:
        """
        :return: The actual operation object which may be any callable.
        """
        return self._wrapped_op

    def __str__(self):
        return '%s: %s' % (self._wrapped_op, self._op_meta_info)

    def __call__(self, *args, monitor: Monitor = Monitor.NONE, **kwargs):
        """
        Perform this operation.

        :param args: the arguments
        :param monitor: an optional progress monitor, which is passed to the
            wrapped callable, if it supports it.
        :param kwargs: the keyword arguments
        :return: the operation output.
        """

        input_values = kwargs

        # process arguments, if any
        num_args = len(args)
        if num_args:
            input_names = self.op_meta_info.input_names
            for position in range(num_args):
                if position >= len(input_names):
                    raise ValueError(
                        "too many inputs given for operation '{}'".format(
                            self.op_meta_info.qualified_name)
                    )
                input_name = self.op_meta_info.input_names[position]
                input_values[input_name] = args[position]

        # set default_value where input values are missing
        self.op_meta_info.set_default_input_values(input_values)

        # validate the input_values using this operation's meta-info
        self.op_meta_info.validate_input_values(
            input_values, validation_exception_class=ValidationError
        )

        if self.op_meta_info.has_monitor:
            # set the monitor only if it is an argument
            input_values[_MONITOR] = monitor

        # call the callable
        return_value = self._wrapped_op(**input_values)

        if self.op_meta_info.has_named_outputs:
            # return_value is expected to be a dictionary-like object
            # set default_value where output values in return_value are missing
            for name, properties in self.op_meta_info.outputs.items():
                if name not in return_value or return_value[name] is None:
                    return_value[name] = properties.get('default_value')
            # validate the return_value using this operation's meta-info
            self.op_meta_info.validate_output_values(return_value)
            # Add history information to outputs
            for name, properties in self.op_meta_info.outputs.items():
                add_history = properties.get('add_history')
                if add_history:
                    return_value[name] = self._add_history(
                        return_value[name], input_values
                    )
        else:
            # return_value is a single value, not a dict
            # set default_value if return_value is missing
            properties = self.op_meta_info.outputs[_RETURN]
            if return_value is None:
                return_value = properties.get('default_value')
            # validate the return_value using this operation's meta-info
            self.op_meta_info.validate_output_values({_RETURN: return_value})
            # Add history information to the output
            add_history = properties.get('add_history')
            if add_history:
                return_value = self._add_history(return_value, input_values)

        return return_value

    def _add_history(self, ds: object, input_dict) -> object:
        """
        Add provenance information about the ESA Climate Toolbox,
        the operation and its inputs to the given output.

        :return: Dataset with history information appended
        """
        op_name = self.op_meta_info.qualified_name
        # There can potentially be different ways to stamp an output depending
        # on its type
        if not isinstance(ds, xr.Dataset):
            raise NotImplementedError(
                'Operation "{}": Adding history information to an output is '
                'currently implemented only for outputs of type '
                '"xarray.Dataset".'.format(op_name))

        # Construct our own dict to stringify, otherwise the full input dataset
        # repr will be found in history.
        input_str = dict()
        for key in input_dict:
            value = input_dict[key]
            if isinstance(value, xr.Dataset):
                # We only show that 'a dataset' was provided, instead of
                # putting the full dataset repr in history
                input_str[key] = type(value)
                continue
            input_str[key] = value

        # Format the stamp
        try:
            op_version = self.op_meta_info.header['version']
        except KeyError:
            raise ValueError(
                'Operation "{}": Could not add history information because the '
                '"version" property is undefined.'.format(op_name))

        stamp = '\nModified with the ESA Climate Toolbox v' \
                + __version__ + ' ' + op_name + ' v' + op_version + \
                ' \nDefault input values: ' + \
                str(self.op_meta_info.inputs) + '\nProvided input values: ' + \
                str(input_str) + '\n'

        # Append the stamp to existing history information or create history
        # attribute if none is found
        if 'history' in ds.attrs:
            if isinstance(ds.attrs['history'], str):
                ds.attrs['history'] = ds.attrs['history'] + stamp
            elif isinstance(ds.attrs['history'], list):
                if isinstance(ds.attrs['history'], dict):
                    ds.attrs['history'][0]['program'] = \
                        ds.attrs['history'][0]['program'] + stamp
                else:
                    ds.attrs['history'].append(stamp)
            else:
                ds.attrs['history'] = stamp
        else:
            ds.attrs['history'] = stamp
        return ds


class OpRegistry:
    """
    An operation registry allows for addition, removal, and retrieval of
    operations.
    """

    def __init__(self):
        self._op_registrations = OrderedDict()

    @property
    def op_registrations(self) -> OrderedDict:
        """
        Get all operation registrations of type
        :py:class:`ect_climate_toolbox.core.op.Operation`.

        :return: a mapping of fully qualified operation names to operation
            registrations
        """
        return OrderedDict(
            sorted(self._op_registrations.items(), key=lambda item: item[0])
        )

    def add_op(self, operation: Callable, fail_if_exists=True,
               replace_if_exists=False) -> Operation:
        """
        Add a new operation registration.

        :param operation: A operation object such as a class or any callable.
        :param fail_if_exists: raise ``ValueError`` if the operation was
            already registered
        :param replace_if_exists: replaces an existing operation
            if *fail_if_exists* is ``False``
        :return: a new or existing
            :py:class:`ect_climate_toolbox.core.op.Operation`
        """
        operation = self._unwrap_operation(operation)
        op_key = self.get_op_key(operation)
        if op_key in self._op_registrations:
            if fail_if_exists:
                raise ValueError(
                    "operation with name '%s' already registered" % op_key
                )
            elif not replace_if_exists:
                return self._op_registrations[op_key]
        op_registration = Operation(operation)
        self._op_registrations[op_key] = op_registration
        return op_registration

    def remove_op(
            self, operation: Callable, fail_if_not_exists=False
    ) -> Optional[Operation]:
        """
        Remove an operation registration.

        :param operation: A fully qualified operation name or operation object
            such as a class or any callable.
        :param fail_if_not_exists: raise ``ValueError`` if no such operation
            was found
        :return: the removed :py:class:`ect_climate_toolbox.core.op.Operation`
            object or ``None`` if *fail_if_not_exists* is ``False``.
        """
        operation = self._unwrap_operation(operation)
        op_key = self.get_op_key(operation)
        if op_key not in self._op_registrations:
            if fail_if_not_exists:
                raise ValueError(
                    "operation with name '%s' not registered" % op_key
                )
            else:
                return None
        return self._op_registrations.pop(op_key)

    def get_op(self, operation, fail_if_not_exists=False) -> Operation:
        """
        Get an operation registration.

        :param operation: A fully qualified operation name or operation object
            such as a class or any callable.
        :param fail_if_not_exists: raise ``ValueError`` if no such operation
            was found
        :return: a :py:class:`ect_climate_toolbox.core.op.Operation` object or
            ``None`` if *fail_if_not_exists* is ``False``.
        """
        operation = self._unwrap_operation(operation)
        op_key = self.get_op_key(operation)
        op_registration = self._op_registrations.get(op_key, None)
        if op_registration is None and fail_if_not_exists:
            raise ValueError("operation with name '%s' not registered" % op_key)
        return op_registration

    # noinspection PyMethodMayBeStatic
    def get_op_key(self, operation: Union[str, Callable]):
        """
        Get a key under which the given operation will be registered.

        :param operation: A fully qualified operation name or a callable object
        :return: The operation key
        """
        try:
            qualified_name = operation.op_meta_info.qualified_name
        except AttributeError:
            if isinstance(operation, str):
                qualified_name = operation
            else:
                operation = self._unwrap_operation(operation)
                qualified_name = object_to_qualified_name(operation)
        if qualified_name.startswith('esa_climate_toolbox.ops.'):
            return qualified_name.rsplit('.', maxsplit=1)[1]
        else:
            return qualified_name

    @classmethod
    def _unwrap_operation(cls, operation):
        if not operation:
            raise ValueError('operation must be given')
        try:
            return operation.wrapped_op
        except AttributeError:
            return operation


class _DefaultOpRegistry(OpRegistry):
    def __repr__(self):
        return 'OP_REGISTRY'


# check (nf) - for more flexibility, REGISTRY may be configured by dependency
# injection
# see Python libs 'pinject' (Google), 'inject', and others

#: The default operation registry of type
#: :py:class:`esa_climate_toolbox.core.op.OpRegistry`.
OP_REGISTRY = _DefaultOpRegistry()


def list_operations(op_registry: OpRegistry = OP_REGISTRY,
                    include_qualified_name: bool = False):
    """
    Lists the operations that are provided by the ESA Climate Toolbox.

    :param op_registry: An optional OpRegistry, in case the default one should
        not be used.
    :param include_qualified_name: If true, a more expressive qualified name
        will be returned along with the method name. Default is false.

    :return: Either a list of the names of operations, or a list of tuples,
        each consisting of the operation name and a qualified name
    """
    op_regs = op_registry.op_registrations

    def _is_op_selected(op_name: str,
                        op_reg,
                        tag_part: str,
                        internal_only: bool,
                        deprecated_only: bool):
        if op_name.startswith('_'):
            # do not list private operations
            return False
        if deprecated_only \
                and not op_reg.op_meta_info.header.get('deprecated'):
            # do not list non-deprecated operations
            # if user wants to see what is deprecated
            return False
        tags = to_list(op_reg.op_meta_info.header.get('tags'))
        if tags:
            # Tagged operations
            if internal_only:
                if 'internal' not in tags:
                    return False
            else:
                if 'internal' in tags:
                    return False
            if tag_part:
                tag_part = tag_part.lower()
                if isinstance(tags, list):
                    return any(tag_part in tag.lower() for tag in tags)
                elif isinstance(tags, str):
                    return tag_part in tags.lower()
        elif internal_only or tag_part:
            # Untagged operations
            return False
        return True
    if include_qualified_name:
        return sorted(
            [(op_name, op_reg.op_meta_info.qualified_name)
             for op_name, op_reg in op_regs.items()
             if _is_op_selected(op_name, op_reg, None, False, False)]
        )
    else:
        return sorted(
            [op_name for op_name, op_reg in op_regs.items()
            if _is_op_selected(op_name, op_reg, None, False, False)]
        )


def get_op_meta_info(
        op_name: str, op_registry: OpRegistry = OP_REGISTRY
) -> Dict:
    """
    Returns meta information about an operation.

    :param op_name: The name of the operation for which meta information shall
        be provided.
    :param op_registry: An optional OpRegistry, in case the default one should
        not be used.

    :return: A dictionary representation of an operator's meta info,
        providing information about input parameters and the expected output.
    """
    op = op_registry.get_op(op_name)
    return op.op_meta_info.to_json_dict()


def get_op(
        op_name: str, op_registry: OpRegistry = OP_REGISTRY
) -> Operation:
    """
    Returns an operation.

    :param op_name: The name of the operation.
    :param op_registry: An optional OpRegistry, in case the default one should
        not be used.

    :return: An operation which may directly be called
    """
    return op_registry.get_op(op_name)


def op(tags=UNDEFINED,
       version=UNDEFINED,
       res_pattern=UNDEFINED,
       deprecated=UNDEFINED,
       registry=OP_REGISTRY,
       **properties):
    """
    ``op`` is a decorator function that registers a Python function or class in
    the default operation registry or the one given by *registry*, if any.
    Any other keywords arguments in *header* are added to the operation's
    meta-information header.
    Classes annotated by this decorator must have callable instances.

    When a function is registered, an introspection is performed. During this
    process, initial operation the meta-information header property
    *description* is derived from the function's docstring.

    If any output of this operation will have its history information
    automatically updated, there should be version information found in the
    operation header. Thus it's always a good idea to add it to all operations::

    @op(version='X.x')

    :param tags: An optional list of string tags.
    :param version: An optional version string.
    :param res_pattern: An optional pattern that will be used to generate the
        names for data resources that are used to hold a reference to the
        objects returned by the operation. Currently, the only pattern variable
        that is supported and that must be present is ``{index}`` which will be
        replaced by an integer number that is guaranteed to produce a unique
        resource name.
    :param deprecated: An optional boolean or a string. If a string is used, it
        should explain why the operation has been deprecated and which new
        operation to use instead. If set to ``True``, the operation's doc-string
        should explain the deprecation.
    :param registry: The operation registry.
    :param properties: Other properties (keyword arguments) that will be added
        to the meta-information of operation.
    """

    def decorator(op_func):
        new_properties = dict(tags=tags,
                              version=version,
                              res_pattern=res_pattern,
                              deprecated=deprecated,
                              **properties)
        op_registration = registry.add_op(op_func, fail_if_exists=False)
        op_registration.op_meta_info.header.update(
            {k: v for k, v in new_properties.items() if v is not UNDEFINED}
        )
        return op_registration

    return decorator


def op_input(input_name: str,
             default_value=UNDEFINED,
             units=UNDEFINED,
             data_type=UNDEFINED,
             nullable=UNDEFINED,
             value_set_source=UNDEFINED,
             value_set=UNDEFINED,
             value_range=UNDEFINED,
             script_lang=UNDEFINED,
             deprecated=UNDEFINED,
             position=UNDEFINED,
             context=UNDEFINED,
             registry=OP_REGISTRY,
             **properties):
    """
    ``op_input`` is a decorator function that provides meta-information for an
    operation input identified by *input_name*. If the decorated function or
    class is not registered as an operation yet, it is added to the default
    operation registry or the one given by *registry*, if any.

    When a function is registered, an introspection is performed. During this
    process, initial operation meta-information input properties are derived
    for each positional and keyword argument named *input_name*:

    ================ ===========================================================
    Derived property Source
    ================ ===========================================================
    *position*       The position of a positional argument, e.g. ``2`` for
                     input ``z`` in ``def f(x, y, z, c=2)``.
    *default_value*  The value of a keyword argument, e.g. ``52.3`` for input
                     ``latitude`` from argument definition
                     ``latitude:float=52.3``
    *data_type*      The type annotation type, e.g. ``float`` for input
                     ``latitude`` from argument definition ``latitude:float``
    ================ ===========================================================

    The derived properties listed above plus any of *value_set*, *value_range*,
    and any key-value pairs in *properties* are added to the input's
    meta-information. A key-value pair in *properties* will always overwrite the
    derived properties listed above.

    :param input_name: The name of an input.
    :param default_value: A default value.
    :param units: The geo-physical units of the input value.
    :param data_type: The data type of the input values.
        If not given, the type of any given, non-None *default_value* is used.
    :param nullable: If ``True``, the value of the input may be ``None``.
        If not given, it will be set to ``True`` if the *default_value* is
        ``None``.
    :param value_set_source: The name of an input, which can be used to
        generate a dynamic value set.
    :param value_set: A sequence of the valid values. Note that all values in
        this sequence must be compatible with *data_type*.
    :param value_range: A sequence specifying the possible range of valid
        values.
    :param script_lang: The programming language for a parameter of data_type
        "str" that provides source code of a script, e.g. "python".
    :param deprecated: An optional boolean or a string. If a string is used, it
        should explain why the input has been deprecated and which new input to
        use instead. If set to ``True``, the input's doc-string should explain
        the deprecation.
    :param position: The zero-based position of an input.
    :param context: If ``True``, the value of the operation input will be a
        dictionary representing the current execution context.  If *context* is
        a string, the value of the operation input will be the result of
        evaluating the string as Python expression with the current execution
        context as local environment. This means, *context* may be an expression
        such as 'value_cache', 'workspace.base_dir', 'step', 'step.id'.
    :param properties: Other properties (keyword arguments) that will be added
        to the meta-information of the named output.
    :param registry: Optional operation registry.
    """

    def decorator(op_func):
        op_registration = registry.add_op(op_func, fail_if_exists=False)
        input_namespace = op_registration.op_meta_info.inputs
        if input_name not in input_namespace:
            input_namespace[input_name] = dict()
        new_properties = dict(data_type=data_type,
                              default_value=default_value,
                              units=units,
                              nullable=nullable,
                              value_set_source=value_set_source,
                              value_set=value_set,
                              value_range=value_range,
                              script_lang=script_lang,
                              deprecated=deprecated,
                              position=position,
                              context=context,
                              **properties)

        input_namespace[input_name].update(
            {k: v for k, v in new_properties.items() if v is not UNDEFINED}
        )
        _adjust_input_properties(input_namespace[input_name])
        return op_registration

    return decorator


def op_output(output_name: str,
              data_type=UNDEFINED,
              deprecated=UNDEFINED,
              registry=OP_REGISTRY,
              **properties):
    """
    ``op_output`` is a decorator function that provides meta-information for an
    operation output identified by *output_name*. If the decorated function or
    class is not registered as an operation yet, it is added to the default
    operation registry or the one given by *registry*, if any.

    If your function does not return multiple named outputs, use the
    :py:func:`op_return` decorator function.
    Note that::

        @op_return(...)
        def my_func(...):
            ...

    if equivalent to::

        @op_output('return', ...)
        def my_func(...):
            ...

    To automatically add information about the ESA Climate Toolbox, its version,
    this operation and its inputs, to this output, set 'add_history' to True::

        @op_output('name', add_history=True)

    Note that the operation should have version information added to it when
    add_history is True::

        @op(version='X.x')

    :param output_name: The name of the output.
    :param data_type: The data type of the output value.
    :param deprecated: An optional boolean or a string. If a string is used, it
        should explain why the output has been deprecated and which new output
        to use instead. If set to ``True``, the output's doc-string should
        explain the deprecation.
    :param properties: Other properties (keyword arguments) that will be added
        to the meta-information of the named output.
    :param registry: Optional operation registry.
    """

    def decorator(op_func):
        op_registration = registry.add_op(op_func, fail_if_exists=False)
        output_namespace = op_registration.op_meta_info.outputs
        if not op_registration.op_meta_info.has_named_outputs:
            # if there is only one entry and it is the 'return' entry,
            # rename it to value of output_name
            output_properties = output_namespace[OpMetaInfo.RETURN_OUTPUT_NAME]
            del output_namespace[OpMetaInfo.RETURN_OUTPUT_NAME]
            output_namespace[output_name] = output_properties
        elif output_name not in output_namespace:
            output_namespace[output_name] = dict()
        new_properties = dict(
            data_type=data_type, deprecated=deprecated, **properties
        )
        output_namespace[output_name].update(
            {k: v for k, v in new_properties.items() if v is not UNDEFINED}
        )
        return op_registration

    return decorator


def op_return(data_type=UNDEFINED,
              registry=OP_REGISTRY,
              **properties):
    """
    ``op_return`` is a decorator function that provides meta-information for a
    single, anonymous operation return value (whose output name is
    ``"return"``). If the decorated function or class is not registered as an
    operation yet, it is added to the default operation registry or the one
    given by *registry*, if any. Any other keywords arguments in *properties*
    are added to the output's meta-information.

    When a function is registered, an introspection is performed. During this
    process, initial operation meta-information output properties are derived
    from the function's return type annotation, that is *data_type* will be,
    e.g., ``float`` if a function is annotated as ``def f(x, y) -> float: ...``.

    The derived *data_type* property and any key-value pairs in *properties* are
    added to the output's meta-information. A key-value pair in *properties*
    will always overwrite a derived *data_type*.

    If your function returns multiple named outputs, use the
    :py:func:`op_output` decorator function.
    Note that::

        @op_return(...)
        def my_func(...):
            ...

    if equivalent to::

        @op_output('return', ...)
        def my_func(...):
            ...

    To automatically add information about the ESA Climate Toolbox, its version,
    this operation and its inputs, to this output, set 'add_history' to True::

        @op_return(add_history=True)

    Note that the operation should have version information added to it when
    add_history is True::

        @op(version='X.x')

    :param data_type: The data type of the return value.
    :param properties: Other properties (keyword arguments)
        that will be added to the meta-information of the return value.
    :param registry: The operation registry.
    """
    return op_output(OpMetaInfo.RETURN_OUTPUT_NAME,
                     data_type=data_type,
                     registry=registry,
                     **properties)


def _adjust_input_properties(input_properties):
    """
    Adjust any undefined input properties that can be derived from other
    defined input properties.
    """

    default_value = input_properties.get('default_value', UNDEFINED)

    # Derive undefined 'nullable' from 'default_value'
    nullable = input_properties.get('nullable', UNDEFINED)
    if nullable is UNDEFINED and default_value is None:
        input_properties['nullable'] = True

    # Derive undefined 'data_type' from 'default_value'
    data_type = input_properties.get('data_type', UNDEFINED)
    if data_type is UNDEFINED and \
            not (default_value is UNDEFINED or default_value is None):
        input_properties['data_type'] = type(default_value)


def new_subprocess_op(op_meta_info: OpMetaInfo,
                      command_pattern: str,
                      run_python: bool = False,
                      cwd: Optional[str] = None,
                      env: Dict[str, str] = None,
                      shell: bool = False,
                      started: Union[str, Callable] = None,
                      progress: Union[str, Callable] = None,
                      done: Union[str, Callable] = None) -> Operation:
    r"""
    Create an operation for a child program run in a new process.

    :param op_meta_info: Meta-information about the resulting operation and
        the operation's inputs and outputs.
    :param command_pattern: A pattern that will be interpolated to obtain the
        actual command to be executed.
        May contain "{input_name}" fields which will be replaced by the actual
        input value converted to text. *input_name* must refer to a valid
        operation input name in *op_meta_info.input* or it must be the value of
        either the "write_to" or "read_from" property of another input's
        property map.
    :param run_python: If True, *command_pattern* refers to a Python script
        which will be executed with the Python interpreter that the Climate
        Toolbox uses.
    :param cwd: Current working directory to run the command line in.
    :param env: Environment variables passed to the shell that executes the
        command line.
    :param shell: Whether to use the shell as the program to execute.
    :param started: Either a callable that receives a text line from the
        executable's stdout and returns a tuple (label, total_work) or a regex
        that must match in order to signal the start of progress monitoring.
        The regex must provide the group names "label" or "total_work" or both,
        e.g. "(?P<label>\w+)" or "(?P<total_work>\d+)"
    :param progress: Either a callable that receives a text line from the
        executable's stdout and returns a tuple (work, msg) or a regex that must
        match in order to signal process.
        The regex must provide group names "work" or "msg" or both, e.g.,
        "(?P<msg>\w+)" or "(?P<work>\d+)"
    :param done: Either a callable that receives a text line a text line from
        the executable's stdout and returns True or False or a regex that must
        match in order to signal the end of progress monitoring.
    :return: The executable wrapped into an operation.
    """

    if started or progress and not op_meta_info.has_monitor:
        op_meta_info = OpMetaInfo(op_meta_info.qualified_name,
                                  has_monitor=True,
                                  inputs=op_meta_info.inputs,
                                  outputs=op_meta_info.outputs,
                                  header=op_meta_info.header)

    # Idea: process special input properties:
    #   - "is_cwd" - an input that provides the current working directory,
    #   must be of type str
    #   - "is_env" - an input that provides environment variables,
    #   must be of type DictLike
    #   - "is_output" - an input that provides the file path of an output,
    #   must be of type str

    def run_executable(**kwargs):

        format_kwargs = {}
        temp_input_files = {}
        temp_output_files = {}

        for name, props in op_meta_info.inputs.items():
            value = kwargs.get(name, props.get('default_value', UNDEFINED))
            if value is not UNDEFINED:
                if 'write_to' in props:
                    new_name = props['write_to']
                    _, file = new_temp_file(suffix='.nc')
                    value.to_netcdf(file)
                    format_kwargs[new_name] = file
                    temp_input_files[name] = file
                else:
                    try:
                        value = value.format()
                    except AttributeError:
                        pass
                    format_kwargs[name] = value

        for name, props in op_meta_info.outputs.items():
            if 'read_from' in props:
                new_name = props['read_from']
                _, file = new_temp_file(suffix='.nc')
                format_kwargs[new_name] = file
                temp_output_files[name] = file

        monitor = None
        if _MONITOR in format_kwargs:
            monitor = format_kwargs.pop(_MONITOR)

        command = command_pattern.format(**format_kwargs)

        stdout_handler = None
        if monitor:
            stdout_handler = ProcessOutputMonitor(
                monitor, label=command, started=started, progress=progress,
                done=done
            )

        if run_python:
            command = '"{}" {}'.format(sys.executable, command)

        exit_code = run_subprocess(
            command, cwd=cwd, env=env, shell=shell,
            stdout_handler=stdout_handler,
            is_cancelled=monitor.is_cancelled if monitor else None
        )

        for file in temp_input_files.values():
            del_temp_file(file)

        return_value = {}
        for name, file in temp_output_files.items():
            return_value[name] = xr.open_dataset(file)

        if not return_value:
            # No output specified, so we return exit code
            return exit_code

        if exit_code:
            # There is output specified, but exit code signals error
            raise ValueError('command [{}] exited with code {}'.format(
                command_pattern, exit_code)
            )

        if len(return_value) == 1 and 'return' in return_value:
            # Single output
            return return_value['return']
        else:
            # Multiple outputs
            return return_value

    run_executable.__name__ = op_meta_info.qualified_name
    run_executable.__doc__ = op_meta_info.header.get('description')

    return Operation(run_executable, op_meta_info=op_meta_info)


def new_expression_op(op_meta_info: OpMetaInfo, expression: str) -> Operation:
    """
    Create an operation that wraps a Python expression.

    :param op_meta_info: Meta-information about the resulting operation and the
        operation's inputs and outputs.
    :param expression: The Python expression. May refer to any name given in
        *op_meta_info.input*.
    :return: The Python expression wrapped into an operation.
    """

    if not op_meta_info:
        raise ValueError('op_meta_info must be given')
    if not expression:
        raise ValueError('expression must be given')

    def eval_expression(**kwargs):
        return safe_eval(expression, local_namespace=kwargs)

    inputs = OrderedDict(op_meta_info.inputs)
    outputs = OrderedDict(op_meta_info.outputs)
    if len(outputs) == 0:
        outputs[_RETURN] = {}
    op_meta_info = OpMetaInfo(op_meta_info.qualified_name,
                              has_monitor=op_meta_info.has_monitor,
                              header=dict(op_meta_info.header),
                              inputs=inputs,
                              outputs=outputs)

    eval_expression.__name__ = op_meta_info.qualified_name
    eval_expression.__doc__ = op_meta_info.header.get('description')

    return Operation(eval_expression, op_meta_info=op_meta_info)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Provides various resampling methods including up- and downsampling.

Components
==========
"""
# http://stackoverflow.com/questions/7075082/what-is-future-in-python-used-for-and-how-when-to-use-it-and-how-it-works

from __future__ import division

import numpy as np
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import resample_in_space

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.types import DatasetLike


#: Interpolation method for upsampling: Take nearest source grid cell,
# even if it is invalid.
US_NEAREST = 10
#: Interpolation method for upsampling:
# Bi-linear interpolation between the 4 nearest source grid cells.
US_LINEAR = 11

#: Aggregation method for downsampling:
# Take first valid source grid cell, ignore contribution areas.
DS_FIRST = 50
#: Aggregation method for downsampling:
# Take last valid source grid cell, ignore contribution areas.
DS_LAST = 51
# DS_MIN = 52
# DS_MAX = 53
#: Aggregation method for downsampling:
# Compute average of all valid source grid cells,
#: with weights given by contribution area.
DS_MEAN = 54
# DS_MEDIAN = 55
#: Aggregation method for downsampling: Compute most frequently seen valid
# source grid cell,with frequency given by contribution area. Note that this
# mode can use an additional keyword argument *mode_rank* which can be used to
# generate the n-th mode. See :py:function:`downsample_2d`.
DS_MODE = 56
#: Aggregation method for downsampling: Compute the biased weighted estimator of
# variance (see https://en.wikipedia.org/wiki/Mean_square_weighted_deviation),
# with weights given by contribution area.
DS_VAR = 57
#: Aggregation method for downsampling: Compute the corresponding standard
# deviation to the biased weighted estimator of variance
#: (see https://en.wikipedia.org/wiki/Mean_square_weighted_deviation),
# with weights given by contribution area.
DS_STD = 58

#: Constant indicating an empty 2-D mask
_NOMASK2D = np.ma.getmaskarray(np.ma.array([[0]], mask=[[0]]))

_EPS = 1e-10


def resample_2d(
        src, w, h, ds_method=DS_MEAN, us_method=US_LINEAR,
        fill_value=None, mode_rank=1, out=None
):
    """
    Resample a 2-D grid to a new resolution.

    :param src: 2-D *ndarray*
    :param w: *int*
        New grid width
    :param h:  *int*
        New grid height
    :param ds_method: one of the *DS_* constants, optional
        Grid cell aggregation method for a possible downsampling
    :param us_method: one of the *US_* constants, optional
        Grid cell interpolation method for a possible upsampling
    :param fill_value: *scalar*, optional
        If ``None``, it is taken from **src** if it is a masked array,
        otherwise from *out* if it is a masked array,
        otherwise numpy's default value is used.
    :param mode_rank: *scalar*, optional
        The rank of the frequency determined by the *ds_method* ``DS_MODE``.
        One (the default) means most frequent value, two means second most
        frequent value, and so forth.
    :param out: 2-D *ndarray*, optional
        Alternate output array in which to place the result. The default is
        *None*; if provided, it must have the same shape as the expected output.
    :return: An resampled version of the *src* array.
    """
    out = _get_out(out, src, (h, w))
    if out is None:
        return src
    mask, use_mask = _get_mask(src)
    fill_value = _get_fill_value(fill_value, src, out)
    return _mask_or_not(
        _resample_2d(
            src, mask, use_mask, ds_method, us_method, fill_value, mode_rank,
            out
        ),
        src, fill_value
    )


def upsample_2d(src, w, h, method=US_LINEAR, fill_value=None, out=None):
    """
    Upsample a 2-D grid to a higher resolution by interpolating original grid
    cells.

    :param src: 2-D *ndarray*
    :param w: *int*
        Grid width, which must be greater than or equal to *src.shape[-1]*
    :param h:  *int*
        Grid height, which must be greater than or equal to *src.shape[-2]*
    :param method: one of the *US_* constants, optional
        Grid cell interpolation method
    :param fill_value: *scalar*, optional
        If ``None``, it is taken from **src** if it is a masked array,
        otherwise from *out* if it is a masked array,
        otherwise numpy's default value is used.
    :param out: 2-D *ndarray*, optional
        Alternate output array in which to place the result. The default is
        *None*; if provided, it must have the same shape as the expected output.
    :return: An upsampled version of the *src* array.
    """
    out = _get_out(out, src, (h, w))
    if out is None:
        return src
    mask, use_mask = _get_mask(src)
    fill_value = _get_fill_value(fill_value, src, out)
    return _mask_or_not(
        _upsample_2d(
            src, mask, use_mask, method, fill_value, out),
        src, fill_value
    )


def downsample_2d(
        src, w, h, method=DS_MEAN, fill_value=None, mode_rank=1, out=None
):
    """
    Downsample a 2-D grid to a lower resolution by aggregating original grid
    cells.

    :param src: 2-D *ndarray*
    :param w: *int*
        Grid width, which must be less than or equal to *src.shape[-1]*
    :param h:  *int*
        Grid height, which must be less than or equal to *src.shape[-2]*
    :param method: one of the *DS_* constants, optional
        Grid cell aggregation method
    :param fill_value: *scalar*, optional
        If ``None``, it is taken from **src** if it is a masked array,
        otherwise from *out* if it is a masked array,
        otherwise numpy's default value is used.
    :param mode_rank: *scalar*, optional
        The rank of the frequency determined by the *method* ``DS_MODE``.
        One (the default) means most frequent value, two means second most
        frequent value, and so forth.
    :param out: 2-D *ndarray*, optional
        Alternate output array in which to place the result. The default is
        *None*; if provided, it must have the same shape as the expected output.
    :return: A downsampled version of the *src* array.
    """
    if method == DS_MODE and mode_rank < 1:
        raise ValueError('mode_rank must be >= 1')
    out = _get_out(out, src, (h, w))
    if out is None:
        return src
    mask, use_mask = _get_mask(src)
    fill_value = _get_fill_value(fill_value, src, out)
    return _mask_or_not(
        _downsample_2d(
            src, mask, use_mask, method, fill_value, mode_rank, out),
        src, fill_value
    )


def _get_out(out, src, shape):
    if out is None:
        return np.zeros(shape, dtype=src.dtype)
    else:
        if out.shape != shape:
            raise ValueError("'shape' and 'out' are incompatible")
        if out.shape == src.shape:
            return None
        return out


def _get_mask(src):
    if isinstance(src, np.ma.MaskedArray):
        mask = np.ma.getmask(src)
        if mask is not np.ma.nomask:
            return mask, True
    return _NOMASK2D, False


def _mask_or_not(out, src, fill_value):
    if isinstance(src, np.ma.MaskedArray):
        if not isinstance(out, np.ma.MaskedArray):
            if np.isfinite(fill_value):
                masked = np.ma.masked_equal(out, fill_value, copy=False)
            else:
                masked = np.ma.masked_invalid(out, copy=False)
            masked.set_fill_value(fill_value)
            return masked
    return out


def _get_fill_value(fill_value, src, out):
    if fill_value is None:
        if isinstance(src, np.ma.MaskedArray):
            fill_value = src.fill_value
        elif isinstance(out, np.ma.MaskedArray):
            fill_value = out.fill_value
        else:
            # use numpy's default fill_value
            fill_value = \
                np.ma.array([0], mask=[False], dtype=src.dtype).fill_value
    return fill_value


# This function will be JIT-compiled by Numba with nopython=True,
# therefore all arg types must be either primitive scalars or numpy arrays.
# Key-value args are not allowed.
#
# disabled jit while numba does not support masked arrays
# https://github.com/numba/numba/issues/1834
# @jit(nopython=False)
def _resample_2d(
        src, mask, use_mask, ds_method, us_method, fill_value, mode_rank, out
):
    src_w = src.shape[-1]
    src_h = src.shape[-2]
    out_w = out.shape[-1]
    out_h = out.shape[-2]

    if out_w < src_w and out_h < src_h:
        return _downsample_2d(
            src, mask, use_mask, ds_method, fill_value, mode_rank, out
        )
    elif out_w < src_w:
        if out_h > src_h:
            temp = np.zeros((src_h, out_w), dtype=src.dtype)
            temp = _downsample_2d(
                src, mask, use_mask, ds_method, fill_value, mode_rank, temp
            )
            # todo - write test & fix: must use mask=np.ma.getmaskarray(temp)
            #  here if use_mask==True
            return _upsample_2d(
                temp, mask, use_mask, us_method, fill_value, out
            )
        else:
            return _downsample_2d(
                src, mask, use_mask, ds_method, fill_value, mode_rank, out
            )
    elif out_h < src_h:
        if out_w > src_w:
            temp = np.zeros((out_h, src_w), dtype=src.dtype)
            temp = _downsample_2d(
                src, mask, use_mask, ds_method, fill_value, mode_rank, temp
            )
            # todo - write test & fix: must use mask=np.ma.getmaskarray(temp)
            #  here if use_mask==True
            return _upsample_2d(
                temp, mask, use_mask, us_method, fill_value, out
            )
        else:
            return _downsample_2d(
                src, mask, use_mask, ds_method, fill_value, mode_rank, out
            )
    elif out_w > src_w or out_h > src_h:
        return _upsample_2d(src, mask, use_mask, us_method, fill_value, out)
    return src


# This function will be JIT-compiled by Numba with nopython=True,
# therefore all arg types must be either primitive scalars or numpy arrays.
# Key-value args are not allowed.
#
# disabled jit while numba does not support masked arrays
# https://github.com/numba/numba/issues/1834
# @jit(nopython=False)
def _upsample_2d(src, mask, use_mask, method, fill_value, out):
    src_w = src.shape[-1]
    src_h = src.shape[-2]
    out_w = out.shape[-1]
    out_h = out.shape[-2]

    if src_w == out_w and src_h == out_h:
        return src

    if out_w < src_w or out_h < src_h:
        raise ValueError("invalid target size")

    if method == US_NEAREST:
        scale_x = src_w / out_w
        scale_y = src_h / out_h
        for out_y in range(out_h):
            src_y = int(scale_y * out_y)
            for out_x in range(out_w):
                src_x = int(scale_x * out_x)
                value = src[src_y, src_x]
                if np.isfinite(value) and not (use_mask and mask[src_y, src_x]):
                    out[out_y, out_x] = value
                else:
                    out[out_y, out_x] = fill_value

    elif method == US_LINEAR:
        scale_x = (src_w - 1.0) / ((out_w - 1.0) if out_w > 1 else 1.0)
        scale_y = (src_h - 1.0) / ((out_h - 1.0) if out_h > 1 else 1.0)
        for out_y in range(out_h):
            src_yf = scale_y * out_y
            src_y0 = int(src_yf)
            wy = src_yf - src_y0
            src_y1 = src_y0 + 1
            if src_y1 >= src_h:
                src_y1 = src_y0
            for out_x in range(out_w):
                src_xf = scale_x * out_x
                src_x0 = int(src_xf)
                wx = src_xf - src_x0
                src_x1 = src_x0 + 1
                if src_x1 >= src_w:
                    src_x1 = src_x0
                v00 = src[src_y0, src_x0]
                v01 = src[src_y0, src_x1]
                v10 = src[src_y1, src_x0]
                v11 = src[src_y1, src_x1]
                if use_mask:
                    v00_ok = np.isfinite(v00) and not mask[src_y0, src_x0]
                    v01_ok = np.isfinite(v01) and not mask[src_y0, src_x1]
                    v10_ok = np.isfinite(v10) and not mask[src_y1, src_x0]
                    v11_ok = np.isfinite(v11) and not mask[src_y1, src_x1]
                else:
                    v00_ok = np.isfinite(v00)
                    v01_ok = np.isfinite(v01)
                    v10_ok = np.isfinite(v10)
                    v11_ok = np.isfinite(v11)
                if v00_ok and v01_ok and v10_ok and v11_ok:
                    ok = True
                    v0 = v00 + wx * (v01 - v00)
                    v1 = v10 + wx * (v11 - v10)
                    value = v0 + wy * (v1 - v0)
                elif wx < 0.5:
                    # NEAREST according to weight
                    if wy < 0.5:
                        ok = v00_ok
                        value = v00
                    else:
                        ok = v10_ok
                        value = v10
                else:
                    # NEAREST according to weight
                    if wy < 0.5:
                        ok = v01_ok
                        value = v01
                    else:
                        ok = v11_ok
                        value = v11
                if ok:
                    out[out_y, out_x] = value
                else:
                    out[out_y, out_x] = fill_value

    else:
        raise ValueError('invalid upsampling method')

    return out


# This function will be JIT-compiled by Numba with nopython=True,
# therefore all arg types must be either primitive scalars or numpy arrays.
# Key-value args are not allowed.
#
# disabled jit while numba does not support masked arrays
# https://github.com/numba/numba/issues/1834
# @jit(nopython=False)
def _downsample_2d(src, mask, use_mask, method, fill_value, mode_rank, out):
    src_w = src.shape[-1]
    src_h = src.shape[-2]
    out_w = out.shape[-1]
    out_h = out.shape[-2]

    if src_w == out_w and src_h == out_h:
        return src

    if out_w > src_w or out_h > src_h:
        raise ValueError("invalid target size")

    scale_x = src_w / out_w
    scale_y = src_h / out_h

    if method == DS_FIRST or method == DS_LAST:
        for out_y in range(out_h):
            src_yf0 = scale_y * out_y
            src_yf1 = src_yf0 + scale_y
            src_y0 = int(src_yf0)
            src_y1 = int(src_yf1)
            if src_y1 == src_yf1 and src_y1 > src_y0:
                src_y1 -= 1
            for out_x in range(out_w):
                src_xf0 = scale_x * out_x
                src_xf1 = src_xf0 + scale_x
                src_x0 = int(src_xf0)
                src_x1 = int(src_xf1)
                if src_x1 == src_xf1 and src_x1 > src_x0:
                    src_x1 -= 1
                done = False
                value = fill_value
                for src_y in range(src_y0, src_y1 + 1):
                    for src_x in range(src_x0, src_x1 + 1):
                        v = src[src_y, src_x]
                        if np.isfinite(v) and \
                                not (use_mask and mask[src_y, src_x]):
                            value = v
                            if method == DS_FIRST:
                                done = True
                                break
                    if done:
                        break
                out[out_y, out_x] = value

    elif method == DS_MODE:
        max_value_count = int(scale_x + 1) * int(scale_y + 1)
        values = np.zeros((max_value_count,), dtype=src.dtype)
        frequencies = np.zeros((max_value_count,), dtype=np.uint32)
        for out_y in range(out_h):
            src_yf0 = scale_y * out_y
            src_yf1 = src_yf0 + scale_y
            src_y0 = int(src_yf0)
            src_y1 = int(src_yf1)
            wy0 = 1.0 - (src_yf0 - src_y0)
            wy1 = src_yf1 - src_y1
            if wy1 < _EPS:
                wy1 = 1.0
                if src_y1 > src_y0:
                    src_y1 -= 1
            for out_x in range(out_w):
                src_xf0 = scale_x * out_x
                src_xf1 = src_xf0 + scale_x
                src_x0 = int(src_xf0)
                src_x1 = int(src_xf1)
                wx0 = 1.0 - (src_xf0 - src_x0)
                wx1 = src_xf1 - src_x1
                if wx1 < _EPS:
                    wx1 = 1.0
                    if src_x1 > src_x0:
                        src_x1 -= 1
                value_count = 0
                for src_y in range(src_y0, src_y1 + 1):
                    wy = wy0 \
                        if (src_y == src_y0) else wy1 \
                        if (src_y == src_y1) else 1.0
                    for src_x in range(src_x0, src_x1 + 1):
                        wx = wx0 if (src_x == src_x0) else wx1 \
                            if (src_x == src_x1) else 1.0
                        v = src[src_y, src_x]
                        if np.isfinite(v) and \
                                not (use_mask and mask[src_y, src_x]):
                            w = wx * wy
                            found = False
                            for i in range(value_count):
                                if v == values[i]:
                                    frequencies[i] += w
                                    found = True
                                    break
                            if not found:
                                values[value_count] = v
                                frequencies[value_count] = w
                                value_count += 1
                w_max = -1.
                value = fill_value
                if mode_rank == 1:
                    for i in range(value_count):
                        w = frequencies[i]
                        if w > w_max:
                            w_max = w
                            value = values[i]
                elif mode_rank <= max_value_count:
                    max_frequencies = np.full(mode_rank, -1.0, dtype=np.float64)
                    indices = np.zeros(mode_rank, dtype=np.int64)
                    for i in range(value_count):
                        w = frequencies[i]
                        for j in range(mode_rank):
                            if w > max_frequencies[j]:
                                max_frequencies[j] = w
                                indices[j] = i
                                break
                    value = values[indices[mode_rank - 1]]

                out[out_y, out_x] = value

    elif method == DS_MEAN:
        for out_y in range(out_h):
            src_yf0 = scale_y * out_y
            src_yf1 = src_yf0 + scale_y
            src_y0 = int(src_yf0)
            src_y1 = int(src_yf1)
            wy0 = 1.0 - (src_yf0 - src_y0)
            wy1 = src_yf1 - src_y1
            if wy1 < _EPS:
                wy1 = 1.0
                if src_y1 > src_y0:
                    src_y1 -= 1
            for out_x in range(out_w):
                src_xf0 = scale_x * out_x
                src_xf1 = src_xf0 + scale_x
                src_x0 = int(src_xf0)
                src_x1 = int(src_xf1)
                wx0 = 1.0 - (src_xf0 - src_x0)
                wx1 = src_xf1 - src_x1
                if wx1 < _EPS:
                    wx1 = 1.0
                    if src_x1 > src_x0:
                        src_x1 -= 1
                v_sum = 0.0
                w_sum = 0.0
                for src_y in range(src_y0, src_y1 + 1):
                    wy = wy0 \
                        if (src_y == src_y0) \
                        else wy1 if (src_y == src_y1) else 1.0
                    for src_x in range(src_x0, src_x1 + 1):
                        wx = wx0 \
                            if (src_x == src_x0) \
                            else wx1 if (src_x == src_x1) else 1.0
                        v = src[src_y, src_x]
                        if np.isfinite(v) and \
                                not (use_mask and mask[src_y, src_x]):
                            w = wx * wy
                            v_sum += w * v
                            w_sum += w
                if w_sum < _EPS:
                    out[out_y, out_x] = fill_value
                else:
                    out[out_y, out_x] = v_sum / w_sum

    elif method == DS_VAR or method == DS_STD:
        for out_y in range(out_h):
            src_yf0 = scale_y * out_y
            src_yf1 = src_yf0 + scale_y
            src_y0 = int(src_yf0)
            src_y1 = int(src_yf1)
            wy0 = 1.0 - (src_yf0 - src_y0)
            wy1 = src_yf1 - src_y1
            if wy1 < _EPS:
                wy1 = 1.0
                if src_y1 > src_y0:
                    src_y1 -= 1
            for out_x in range(out_w):
                src_xf0 = scale_x * out_x
                src_xf1 = src_xf0 + scale_x
                src_x0 = int(src_xf0)
                src_x1 = int(src_xf1)
                wx0 = 1.0 - (src_xf0 - src_x0)
                wx1 = src_xf1 - src_x1
                if wx1 < _EPS:
                    wx1 = 1.0
                    if src_x1 > src_x0:
                        src_x1 -= 1
                v_sum = 0.0
                w_sum = 0.0
                wv_sum = 0.0
                wvv_sum = 0.0
                for src_y in range(src_y0, src_y1 + 1):
                    wy = wy0 \
                        if (src_y == src_y0) \
                        else wy1 if (src_y == src_y1) else 1.0
                    for src_x in range(src_x0, src_x1 + 1):
                        wx = wx0 \
                            if (src_x == src_x0) \
                            else wx1 if (src_x == src_x1) else 1.0
                        v = src[src_y, src_x]
                        if np.isfinite(v) and \
                                not (use_mask and mask[src_y, src_x]):
                            w = wx * wy
                            v_sum += v
                            w_sum += w
                            wv_sum += w * v
                            wvv_sum += w * v * v
                if w_sum < _EPS:
                    out[out_y, out_x] = fill_value
                else:
                    out[out_y, out_x] = \
                        (wvv_sum * w_sum - wv_sum * wv_sum) / w_sum / w_sum
        if method == DS_STD:
            out = np.sqrt(out).astype(out.dtype)
    else:
        raise ValueError('invalid downsampling method')

    return out


@op(tags=['geometric', 'resampling'])
@op_input('ds', data_type=DatasetLike)
@op_input('x_res', data_type=float)
@op_input('y_res', data_type=float)
@op_input('upsampling_float',
          value_set=['nearest_neighbor', 'bilinear', '2nd-order spline',
                     'cubic', '4th-order spline', '5th-order spline'],
          default_value='bilinear')
@op_input('upsampling_int',
          value_set=['nearest_neighbor', 'bilinear', '2nd-order spline',
                     'cubic', '4th-order spline', '5th-order spline'],
          default_value='nearest_neighbor')
@op_input('downsampling_float',
          value_set=['nearest_neighbor', 'mean', 'min', 'max'],
          default_value='mean')
@op_input('downsampling_int',
          value_set=['nearest_neighbor', 'mean', 'min', 'max'],
          default_value='nearest_neighbor')
def resample(
        ds: DatasetLike.TYPE,
        x_res: float,
        y_res: float,
        upsampling_float: str = 'bilinear',
        upsampling_int: str = 'nearest_neighbor',
        downsampling_float: str = 'mean',
        downsampling_int: str = 'nearest_neighbor',
) -> xr.Dataset:
    """
    Resample a dataset to the provided x- and y-resolution. The resolution must
    be given in the units of the CRS.
    It can be set which method to use to upsample integer or float variables
    (in case the new resolution is finer than the old one) or to downsample
    them (in case the new resolution is coarser).

    :param ds: The input dataset.
    :param x_res: The resolution in x-direction.
    :param y_res: The resolution in y-direction.
    :param upsampling_float: The upsampling method to be used for float values.
        This value is only used when the new resolution is finer than the
        previous one. Allowed values are 'nearest_neighbor', 'bilinear',
        '2nd-order spline', 'cubic', '4th-order spline', and '5th-order spline'.
        The default is 'bilinear'.
    :param upsampling_int: The upsampling method to be used for integer and
        boolean values.
        This value is only used when the new resolution is finer than the
        previous one. Allowed values are 'nearest_neighbor', 'bilinear',
        '2nd-order spline', 'cubic', '4th-order spline', and '5th-order spline'.
        The default is 'nearest_neighbor'.
    :param downsampling_float: The downsampling method to be used for float
        values.
        This value is only used when the new resolution is coarser than the
        previous one. Allowed values are 'nearest_neighbor', 'mean', 'min',
        and 'max'.
        The default is 'mean'.
    :param downsampling_int: The downsampling method to be used for integer and
        boolean values.
        This value is only used when the new resolution is coarser than the
        previous one. Allowed values are 'nearest_neighbor', 'mean', 'min',
        and 'max'.
        The default is 'nearest_neighbor'.
    :return: A new dataset resampled to the new resolutions.
    """

    ds = DatasetLike.convert(ds)
    source_gm = GridMapping.from_dataset(ds)
    x_scale = source_gm.x_res / x_res
    y_scale = source_gm.y_res / y_res

    target_gm = source_gm.scale(xy_scale=(x_scale, y_scale))

    var_configs = dict()

    upsampling_values = ['nearest_neighbor', 'bilinear', '2nd-order spline',
                         'cubic', '4th-order spline', '5th-order spline']
    downsampling_values = {
        'nearest_neighbor': None,
        'mean': 'np.nanmean',
        'min': 'np.nanmin',
        'max': 'np.nanmax'
    }

    try:
        usf = upsampling_values.index(upsampling_float)
    except ValueError:
        raise ValueError(f'Unknown upsampling value "{upsampling_float}"')
    try:
        usi = upsampling_values.index(upsampling_int)
    except ValueError:
        raise ValueError(f'Unknown upsampling value "{upsampling_int}"')
    if downsampling_float not in downsampling_values.keys():
        raise ValueError(f'Unknown downsampling value "{downsampling_float}"')
    if downsampling_int not in downsampling_values.keys():
        raise ValueError(f'Unknown downsampling value "{downsampling_int}"')
    dsf = downsampling_values[downsampling_float]
    dsi = downsampling_values[downsampling_int]

    for k, var in ds.variables.items():
        if np.issubdtype(var.dtype, np.integer) \
                or np.issubdtype(var.dtype, bool):
            var_config = dict(
                spline_order=usi,
                aggregator=dsi,
                recover_nan=False,
            )
        else:
            var_config = dict(
                spline_order=usf,
                aggregator=dsf,
                recover_nan=True,
            )
        var_configs[k] = var_config

    return resample_in_space(
        dataset=ds,
        source_gm=source_gm,
        target_gm=target_gm,
        var_configs=var_configs
    )



from unittest import TestCase

from esa_climate_toolbox.util.im.cmaps import get_cmaps


class CmapsTest(TestCase):

    def test_get_cmaps_returns_singleton(self):
        cmaps = get_cmaps()
        self.assertIs(cmaps, get_cmaps())
        self.assertIs(cmaps, get_cmaps())

    def test_get_cmaps_returns_equal_size_recs(self):
        cmaps = get_cmaps()
        rec_len = len(cmaps[0])
        self.assertEqual(rec_len, 3)
        for cmap in cmaps:
            self.assertEqual(len(cmap), rec_len)

    def test_get_cmaps_categories(self):
        cmaps = get_cmaps()
        self.assertGreaterEqual(len(cmaps), 6)
        self.assertEqual(cmaps[0][0], 'Perceptually Uniform Sequential')
        self.assertEqual(cmaps[1][0], 'Sequential 1')
        self.assertEqual(cmaps[2][0], 'Sequential 2')
        self.assertEqual(cmaps[3][0], 'Diverging')
        self.assertEqual(cmaps[4][0], 'Qualitative')
        self.assertEqual(cmaps[5][0], 'Miscellaneous')

    def test_get_cmaps_category_descr(self):
        cmaps = get_cmaps()
        self.assertEqual(cmaps[0][1],
                         'For many applications, a perceptually uniform colormap is '
                         'the best choice - one in which equal steps in data are '
                         'perceived as equal steps in the color space')

    def test_get_cmaps_category_tuples(self):
        cmaps = get_cmaps()
        category_tuple = cmaps[0][2]
        self.assertEqual(len(category_tuple), 4)
        self.assertEqual(category_tuple[0][0], 'viridis')
        self.assertEqual(category_tuple[0][1],
                         'iVBORw0KGgoAAAANSUhEUgAAAQAAAAACCAYAAAC3zQLZAAAAzklEQVR4nO2T'
                         'QZLFIAhEX7dXmyPM/Y8SZwEqMcnU3/9QZTU8GszC6Ee/HQlk5FAsJIENqVGv/'
                         'piZ3uqf3nX6Vtd+l8D8UwNOLhZL3+BLh796OXvMdWaqtrrqnZ/tjvuZT/'
                         '0XxnN/5f25z9X7tIMTKzV7/5yrME3NHoPlUzvplgOevOcz6ZO5eCqzOmark1'
                         'nHDQveHuuYaazZkTcdmE110HJu6doR3tgfPHyL51zNc0fd2xjf0vPukUPL36'
                         'YBTcpcWArFyY0RTca88cYbXxt/gUOJC8yRF1kAAAAASUVORK5CYII=')

        self.assertEqual(category_tuple[1][0], 'inferno')
        self.assertEqual(category_tuple[2][0], 'plasma')
        self.assertEqual(category_tuple[3][0], 'magma')



"""
Tests for timeseries operations
"""

from unittest import TestCase

import numpy as np
import xarray as xr

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.util.misc import object_to_qualified_name

from esa_climate_toolbox.ops.timeseries import tseries_mean
from esa_climate_toolbox.ops.timeseries import tseries_point


def assertDatasetEqual(expected, actual):
    # this method is functionally equivalent to
    # `assert expected == actual`, but it checks each aspect of
    # equality separately for easier debugging
    assert expected.equals(actual), (expected, actual)


class TimeSeriesPoint(TestCase):

    def test_tseries_point(self):
        # Test general functionality
        dataset = xr.Dataset({
            'abs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'bbs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})

        actual = tseries_point(dataset, point='10, 5', var='*bs')
        expected = xr.Dataset({
            'abs': (['time'], np.ones(6)),
            'bbs': (['time'], np.ones(6)),
            'lat': 22.5,
            'lon': 22.5,
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        expected = expected.set_coords(['lat', 'lon'])
        assertDatasetEqual(expected, actual)

        actual = tseries_point(dataset, point=(10, 5), var='')
        assertDatasetEqual(expected, actual)

    def registered(self):
        """
        Test tseries_point as a registered operation
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(tseries_point))
        dataset = xr.Dataset({
            'abs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'bbs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})

        actual = reg_op(ds=dataset, point='10, 5', var='*bs')
        expected = xr.Dataset({
            'abs': (['time'], np.ones(6)),
            'bbs': (['time'], np.ones(6)),
            'lat': 22.5,
            'lon': 22.5,
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        expected = expected.set_coords(['lat', 'lon'])
        assertDatasetEqual(expected, actual)


class TimeSeriesMean(TestCase):

    def test_tseries_mean(self):
        # Test general functionality
        dataset = xr.Dataset({
            'abs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'bbs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        actual = tseries_mean(dataset, var='*bs')
        expected = xr.Dataset({
            'abs_mean': (['time'], np.ones([6])),
            'bbs_mean': (['time'], np.ones([6])),
            'abs_std': (['time'], np.zeros([6])),
            'bbs_std': (['time'], np.zeros([6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        assertDatasetEqual(expected, actual)

        actual = tseries_mean(dataset, var='')
        assertDatasetEqual(actual, expected)

    def registered(self):
        """
        Test tseries_point as a registered operation
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(tseries_mean))
        dataset = xr.Dataset({
            'abs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'bbs': (['lat', 'lon', 'time'], np.ones([4, 8, 6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        actual = reg_op(ds=dataset, var='*bs')
        expected = xr.Dataset({
            'abs': (['time'], np.ones([6])),
            'bbs': (['time'], np.ones([6])),
            'abs_std': (['time'], np.zeros([6])),
            'bbs_std': (['time'], np.zeros([6])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': ['2000-01-01', '2000-02-01', '2000-03-01', '2000-04-01',
                     '2000-05-01', '2000-06-01']})
        assertDatasetEqual(expected, actual)



import sys

import xarray as xr

args = sys.argv
if len(args) < 1 + 3:
    print("usage: filterds.py <in_file> <out_file> <var> ...", file=sys.stderr)
    exit(1)

in_file = args[1]
out_file = args[2]
vars = args[3:]

ds = xr.open_dataset(in_file)
ds = ds.drop_vars(vars)
ds.to_netcdf(out_file)



from unittest import TestCase

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry
import shapely.wkt
from shapely.geometry import Point

from esa_climate_toolbox.core.types import GeoDataFrameProxy
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.ops.data_frame import aggregate_statistics
from esa_climate_toolbox.ops.data_frame import data_frame_max
from esa_climate_toolbox.ops.data_frame import data_frame_min
from esa_climate_toolbox.ops.data_frame import data_frame_subset
from esa_climate_toolbox.ops.data_frame import find_closest
from esa_climate_toolbox.ops.data_frame import great_circle_distance
from esa_climate_toolbox.ops.data_frame import query

test_point = 'POINT (597842.4375881671 5519903.13366397)'

test_poly_4326 = 'POLYGON ((-80 -40, -70 -40, ' \
                 '-70 -45, -80 -45, ' \
                 '-80 -40))'


class TestDataFrameOps(TestCase):
    df = pd.DataFrame({'A': [1, 2, 3, 4, 5, 6],
                       'B': ['a', 'b', 'c', 'x', 'y', 'z'],
                       'C': [False, False, True, False, True, True],
                       'D': [0.4, 0.5, 0.3, 0.3, 0.1, 0.4]})

    gdf = gpd.GeoDataFrame({'A': [1, 2, 3, 4, 5, 6],
                            'B': ['a', 'b', 'c', 'x', 'y', 'z'],
                            'C': [False, False, True, False, True, True],
                            'D': [0.4, 0.5, 0.3, 0.3, 0.1, 0.4],
                            'geometry': gpd.GeoSeries([
                                shapely.wkt.loads('POINT(10 10)'),
                                shapely.wkt.loads('POINT(10 20)'),
                                shapely.wkt.loads('POINT(10 30)'),
                                shapely.wkt.loads('POINT(20 30)'),
                                shapely.wkt.loads('POINT(20 20)'),
                                shapely.wkt.loads('POINT(20 10)'),
                            ])})

    gdf_32718 = gpd.GeoDataFrame({'A': [1]},
                                 crs={'init': 'epsg:32718'},
                                 geometry=[shapely.wkt.loads(test_point)])

    test_region_4326 = shapely.wkt.loads(test_poly_4326)

    gdfp = GeoDataFrameProxy.from_features(gdf.__geo_interface__['features'])

    def test_data_frame_min(self):
        df2 = data_frame_min(TestDataFrameOps.df, 'D')
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertEqual(len(df2), 1)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D'])
        self.assertEqual(df2.iloc[0, 0], 5)
        self.assertEqual(df2.iloc[0, 1], 'y')
        self.assertEqual(df2.iloc[0, 2], True)
        self.assertEqual(df2.iloc[0, 3], 0.1)

    def test_data_frame_max(self):
        df2 = data_frame_max(TestDataFrameOps.df, 'D')
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertEqual(len(df2), 1)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D'])
        self.assertEqual(df2.iloc[0, 0], 2)
        self.assertEqual(df2.iloc[0, 1], 'b')
        self.assertEqual(df2.iloc[0, 2], False)
        self.assertEqual(df2.iloc[0, 3], 0.5)

    def test_query(self):
        df2 = query(TestDataFrameOps.df, "D >= 0.4 and B != 'b'")
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertEqual(len(df2), 2)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D'])
        self.assertEqual(df2.iloc[0, 0], 1)
        self.assertEqual(df2.iloc[0, 1], 'a')
        self.assertEqual(df2.iloc[0, 2], False)
        self.assertEqual(df2.iloc[0, 3], 0.4)
        self.assertEqual(df2.iloc[1, 0], 6)
        self.assertEqual(df2.iloc[1, 1], 'z')
        self.assertEqual(df2.iloc[1, 2], True)
        self.assertEqual(df2.iloc[1, 3], 0.4)

    def test_data_frame_query_with_geom(self):
        gdf = TestDataFrameOps.gdf

        df2 = query(gdf, "not C and @almost_equals('10,10')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        df2 = query(gdf, "not C and @contains('10,10')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        df2 = query(gdf, "not C and @crosses('10,10')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 0)
        df2 = query(gdf, "not C and @disjoint('10,10')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 2)
        df2 = query(gdf, "not C and @intersects('19, 9, 21, 31')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        df2 = query(gdf, "not C and @touches('10, 10, 20, 30')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 3)
        df2 = query(gdf, "@within('19, 9, 21, 31')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 3)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D', 'geometry'])
        self.assertEqual(df2.iloc[0, 0], 4)
        self.assertEqual(df2.iloc[1, 0], 5)
        self.assertEqual(df2.iloc[2, 0], 6)
        self.assertEqual(df2.iloc[0, 1], 'x')
        self.assertEqual(df2.iloc[1, 1], 'y')
        self.assertEqual(df2.iloc[2, 1], 'z')
        self.assertEqual(df2.iloc[0, 2], False)
        self.assertEqual(df2.iloc[1, 2], True)
        self.assertEqual(df2.iloc[2, 2], True)
        self.assertEqual(df2.iloc[0, 3], 0.3)
        self.assertEqual(df2.iloc[1, 3], 0.1)
        self.assertEqual(df2.iloc[2, 3], 0.4)
        df2 = query(gdf, "not C and @within('19, 9, 21, 31')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D', 'geometry'])
        self.assertEqual(df2.iloc[0, 0], 4)
        self.assertEqual(df2.iloc[0, 1], 'x')
        self.assertEqual(df2.iloc[0, 2], False)
        self.assertEqual(df2.iloc[0, 3], 0.3)
        df2 = query(gdf, "not C and @within(@from_wkt('19, 9, 21, 31'))")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        self.assertEqual(list(df2.columns), ['A', 'B', 'C', 'D', 'geometry'])
        self.assertEqual(df2.iloc[0, 0], 4)
        self.assertEqual(df2.iloc[0, 1], 'x')
        self.assertEqual(df2.iloc[0, 2], False)
        self.assertEqual(df2.iloc[0, 3], 0.3)

    def test_data_frame_subset(self):
        df2 = data_frame_subset(TestDataFrameOps.gdf,
                                region='POLYGON((-10 0, 25 0, 25 30, -10 0))')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 3)
        self.assertIn('A', df2)
        self.assertIn('B', df2)
        self.assertIn('C', df2)
        self.assertIn('D', df2)
        self.assertIn('geometry', df2)

        df2 = data_frame_subset(TestDataFrameOps.gdf,
                                var_names="A,C",
                                region='POLYGON((-10 0, 25 0, 25 30, -10 0))')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 3)
        self.assertIn('A', df2)
        self.assertNotIn('B', df2)
        self.assertIn('C', df2)
        self.assertNotIn('D', df2)
        self.assertIn('geometry', df2)

        df2 = data_frame_subset(TestDataFrameOps.gdf,
                                var_names="A,C",
                                region='POLYGON((30 30, 40 30, 40 40, 30 30))')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 0)

    def test_data_frame_failures(self):
        df2 = query(TestDataFrameOps.gdf_32718, "@within('" + test_poly_4326 + "')")
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)

        df2 = data_frame_subset(TestDataFrameOps.gdf_32718,
                                var_names='A',
                                region=TestDataFrameOps.test_region_4326)
        self.assertEqual(len(df2), 1)

    def test_data_frame_find_closest(self):
        df2 = find_closest(TestDataFrameOps.gdf, 'POINT(20 30)', dist_col_name='dist')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 1)
        self.assertIn('A', df2)
        self.assertIn('B', df2)
        self.assertIn('C', df2)
        self.assertIn('D', df2)
        self.assertIn('geometry', df2)
        self.assertIn('dist', df2)
        self.assertEqual(1, len(df2['dist']))
        self.assertEqual(0.0, df2['dist'].iloc[0])
        self.assertEqual(shapely.wkt.loads('POINT(20 30)'), df2['geometry'].iloc[0])

        df2 = find_closest(TestDataFrameOps.gdf, 'POINT(21 28)',
                           max_results=3, dist_col_name='dist')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 3)
        self.assertIn('A', df2)
        self.assertIn('B', df2)
        self.assertIn('C', df2)
        self.assertIn('D', df2)
        self.assertIn('geometry', df2)
        self.assertIn('dist', df2)
        np.testing.assert_approx_equal(2.1828435, df2['dist'].iloc[0])
        np.testing.assert_approx_equal(8.0518568, df2['dist'].iloc[1])
        np.testing.assert_approx_equal(9.8221713, df2['dist'].iloc[2])
        self.assertEqual(shapely.wkt.loads('POINT(20 30)'), df2['geometry'].iloc[0])
        self.assertEqual(shapely.wkt.loads('POINT(20 20)'), df2['geometry'].iloc[1])
        self.assertEqual(shapely.wkt.loads('POINT(10 30)'), df2['geometry'].iloc[2])

        df2 = find_closest(TestDataFrameOps.gdf, 'POINT(21 28)',
                           max_dist=9.0, max_results=3, dist_col_name='dist')
        self.assertIsInstance(df2, gpd.GeoDataFrame)
        self.assertEqual(len(df2), 2)
        self.assertIn('A', df2)
        self.assertIn('B', df2)
        self.assertIn('C', df2)
        self.assertIn('D', df2)
        self.assertIn('geometry', df2)
        self.assertIn('dist', df2)
        np.testing.assert_approx_equal(2.1828435, df2['dist'].iloc[0])
        np.testing.assert_approx_equal(8.0518568, df2['dist'].iloc[1])
        self.assertEqual(shapely.wkt.loads('POINT(20 30)'), df2['geometry'].iloc[0])
        self.assertEqual(shapely.wkt.loads('POINT(20 20)'), df2['geometry'].iloc[1])

    def test_aggregate_statistics(self):
        # Generate mock data
        data = {'name': ['A', 'B', 'C'],
                'lat': [45, 46, 47.5],
                'lon': [-120, -121.2, -122.9]}

        df = pd.DataFrame(data)
        # needs to be a copy
        gdf_empty_geo = gpd.GeoDataFrame(df.copy())
        gdf = gpd.GeoDataFrame(
            df.copy(), geometry=gpd.points_from_xy(x=df["lon"], y=df['lat'])
        )

        var_names_not_agg = 'name, lat, lon'
        var_names_not_in = 'asdc, lat, lon'
        var_names_valid = ['lat', 'lon']
        aggregations = ["count", "mean", "median", "sum", "std", "min", "max"]

        # Assert that a Validation exception is thrown if the df is None
        with self.assertRaises(ValidationError):
            aggregate_statistics(df=None)

        # Assert that a Validation exception is thrown if the var_names contain
        # non-existing fields in the df
        with self.assertRaises(ValidationError):
            aggregate_statistics(df=df, var_names=var_names_not_in)

        # Assert that a Validation exception is thrown if the var_names contain
        # non-aggregatable fields
        with self.assertRaises(ValidationError):
            aggregate_statistics(df=df, var_names=var_names_not_agg)

        # Assert that a Validation exception is thrown if the GeoDataFrame does
        # not have a geometry
        with self.assertRaises(ValidationError):
            aggregate_statistics(df=gdf_empty_geo, var_names=None)

        with self.assertRaises(ValidationError):
            aggregate_statistics(df=gdf_empty_geo, var_names='lat')

        # assert that a input and output types for df are the same
        rdf = aggregate_statistics(df=gdf, var_names=var_names_valid)
        self.assertEqual(len(rdf), 1)

        # assert that columns are returned if var_names = None for a DataFrame
        rdf = aggregate_statistics(df=df, var_names=None)
        self.assertEqual(len(rdf.columns), len(aggregations) * len(var_names_valid))

        # assert that columns are returned if var_names = None for a GeoDataFrame
        rdf = aggregate_statistics(df=gdf, var_names=None, aggregate_geometry=True)
        self.assertEqual(len(rdf.columns), len(aggregations) * len(var_names_valid) + 1)

        # assert that geometry union is created
        rdf = aggregate_statistics(
            df=gdf, var_names=var_names_valid, aggregate_geometry=True
        )
        self.assertIsNotNone(rdf.geometry)


class GreatCircleDistanceTest(TestCase):
    def test_great_circle_distance(self):
        dist = great_circle_distance(Point(20, 20), Point(20, 20))
        self.assertIsNotNone(dist)
        np.testing.assert_approx_equal(0.0, dist)
        dist = great_circle_distance(Point(20, 0), Point(20, 30))
        np.testing.assert_approx_equal(30.0, dist)
        dist = great_circle_distance(Point(-20, 0), Point(20, 0))
        np.testing.assert_approx_equal(40.0, dist)
        dist = great_circle_distance(Point(-155, 0), Point(155, 0))
        np.testing.assert_approx_equal(50.0, dist)
        dist = great_circle_distance(Point(0, 0), Point(0, 90))
        np.testing.assert_approx_equal(90.0, dist)
        dist = great_circle_distance(Point(0, -90), Point(0, 90))
        np.testing.assert_approx_equal(180.0, dist)
        dist = great_circle_distance(Point(0, 180), Point(0, 0))
        np.testing.assert_approx_equal(180.0, dist)
        dist = great_circle_distance(Point(0, 0), Point(1, 1))
        np.testing.assert_approx_equal(1.4141777, dist)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Outlier detection operations

Functions
=========
"""
import fnmatch
import xarray as xr
import numpy as np

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import VarNamesLike
from esa_climate_toolbox.util.monitor import Monitor
from esa_climate_toolbox import __version__


@op(tags=['filter'], version='1.0')
@op_input('ds', data_type=DatasetLike)
@op_input('var', value_set_source='ds', data_type=VarNamesLike)
@op_return(add_history=True)
def detect_outliers(ds: xr.Dataset,
                    var: VarNamesLike.TYPE,
                    threshold_low: float = 0.05,
                    threshold_high: float = 0.95,
                    quantiles: bool = True,
                    mask: bool = False,
                    monitor: Monitor = Monitor.NONE) -> xr.Dataset:
    """
    Detect outliers in the given Dataset.

    When mask=True the input dataset should not contain nan values, otherwise
    all existing nan values will be marked as 'outliers' in the mask data array
    added to the output dataset.

    :param ds: The dataset or dataframe for which to do outlier detection
    :param var: Variable or variables in the dataset to which to do outlier
        detection. Note that when multiple variables are selected, absolute
        threshold values might not make much sense. Wild cards can be used to
        select multiple variables matching a pattern.
    :param threshold_low: Values less or equal to this will be removed/masked
    :param threshold_high: Values greater or equal to this will be
        removed/masked
    :param quantiles: If True, threshold values are treated as quantiles,
        otherwise as absolute values.
    :param mask: If True, an ancillary variable containing flag values for
        outliers will be added to the dataset. Otherwise, outliers will be
        replaced with nan directly in the data variables.
    :param monitor: A progress monitor.
    :return: The dataset with outliers masked or replaced with nan
    """

    ds = DatasetLike.convert(ds)
    # Create a list of variable names on which to perform outlier detection
    # based on the input comma separated list that can contain wildcards
    var_patterns = VarNamesLike.convert(var)
    all_vars = list(ds.data_vars.keys())
    variables = list()
    for pattern in var_patterns:
        leave = fnmatch.filter(all_vars, pattern)
        variables = variables + leave

    # For each array in the dataset for which we should detect outliers, detect
    # outliers
    ret_ds = ds.copy()
    with monitor.starting("detect_outliers", total_work=len(variables) * 3):
        for var_name in variables:
            if quantiles:
                # Get threshold values
                with monitor.child(1).observing("quantile low"):
                    threshold_low = ret_ds[var_name].quantile(threshold_low)
                with monitor.child(1).observing("quantile high"):
                    threshold_high = ret_ds[var_name].quantile(threshold_high)
            else:
                monitor.progress(2)
            # If not mask, put nans in the data arrays for min/max outliers
            if not mask:
                arr = ret_ds[var_name]
                attrs = arr.attrs
                ret_ds[var_name] = arr.where((arr > threshold_low) &
                                             (arr < threshold_high))
                ret_ds[var_name].attrs = attrs
            else:
                # Create and add a data variable containing the mask for this
                # data variable
                _mask_outliers(ret_ds, var_name, threshold_low, threshold_high)
            monitor.progress(1)

    return ret_ds


def _mask_outliers(ds: xr.Dataset, var_name: str, threshold_low: float,
                   threshold_high: float):
    """
    Create a mask data array for the given variable of the dataset and given
    absolute threshold values. Add the mask data array as an ancillary data
    array to the original array as per CF conventions.

    For explanation about the relevant attributes, see::
    http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#flags
    http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#ancillary-data

    :param ds: The dataset (will be mutated)
    :param var_name: variable name
    :param threshold_low: absolute threshold bottom value
    :param threshold_high: absolute threshold top value
    """
    arr = ds[var_name]

    # Create a boolean mask where True denotes an outlier, convert it to 8-bit
    # integer dtype, as to_netcdf will complain about a boolean dtype
    mask = arr.where((arr > threshold_low) & (arr < threshold_high))
    mask = mask.isnull()
    mask = mask.astype('i1')

    # According to CF conventions, the actual variable name in the netCDF can
    # be whatever, but appending things after an underscore is a reasonable
    # convention
    mask_name = var_name + '_outlier_mask'

    # Set the flag data array attributes as per CF conventions
    try:
        mask.attrs['long_name'] = arr.attrs['long_name'] + ' outlier mask'
    except KeyError:
        # The dataset is not CF compliant, add the attribute anyway
        mask.attrs['long_name'] = 'Outlier mask'
    try:
        mask.attrs['standard_name'] = \
            arr.attrs['standard_name'] + ' status_flag'
    except KeyError:
        # The dataset is not CF compliant, add the attribute anyway
        mask.attrs['standard_name'] = 'status_flag'
    mask.attrs['_FillValue'] = 0
    mask.attrs['valid_range'] = np.array([1.0, 1.0], dtype='i1')
    mask.attrs['flag_values'] = np.array([1], dtype='i1')
    mask.attrs['flag_meanings'] = "is_outlier"
    mask.attrs['source'] = "ESA Climate Toolbox v" + __version__

    # Add the mask array to the dataset
    ds[mask_name] = mask

    # Create an ancillary variable link between the parent data array and the
    # mask array
    try:
        anc_var = ds[var_name].attrs['ancillary_variables']
    except KeyError:
        # No ancillary variables associated with this variable yet
        anc_var = ''
    ds[var_name].attrs['ancillary_variables'] = anc_var + ' ' + mask_name



from unittest import TestCase

from esa_climate_toolbox.util.misc import fetch_std_streams
from esa_climate_toolbox.util.monitor import ChildMonitor
from esa_climate_toolbox.util.monitor import ConsoleMonitor
from esa_climate_toolbox.util.monitor import Monitor


class NullMonitorTest(TestCase):
    def test_NONE(self):
        self.assertIsNotNone(Monitor.NONE)
        self.assertEqual(repr(Monitor.NONE), 'Monitor.NONE')

    def test_child_monitor(self):
        self.assertIs(Monitor.NONE.child(work=10), Monitor.NONE)

    def test_cancel(self):
        m = Monitor.NONE
        m.start('task A', total_work=10)
        self.assertFalse(m.is_cancelled())
        m.cancel()
        self.assertFalse(m.is_cancelled())


class ConsoleMonitorTest(TestCase):
    def test_console_monitor_wo_progress_bar(self):
        m = ConsoleMonitor()
        with fetch_std_streams() as (stdout, stderr):
            m.start('task A', total_work=10)
            m.progress(work=1)
            m.progress(work=5, msg='phase 1')
            m.progress(msg='phase 2')
            m.progress(work=4)
            m.progress()
            m.done()
        actual_stdout = stdout.getvalue()
        expected_stdout = 'task A: started\n' + \
                          'task A:  10% \n' + \
                          'task A:  60% phase 1\n' + \
                          'task A: phase 2\n' + \
                          'task A: 100% \n' + \
                          'task A: in progress...\n' + \
                          'task A: done\n'

        self.assertEqual(actual_stdout, expected_stdout)
        self.assertEqual(stderr.getvalue(), '')

    def test_console_monitor_with_progress_bar(self):
        m = ConsoleMonitor(progress_bar_size=10)
        with fetch_std_streams() as (stdout, stderr):
            m.start('task A', total_work=10)
            m.progress(work=1)
            m.progress(work=5, msg='phase 1')
            m.progress(msg='phase 2')
            m.progress(work=4)
            m.progress()
            m.done()
        actual_stdout = stdout.getvalue()
        expected_stdout = 'task A: started\n' + \
                          'task A: [#---------]  10% \n' + \
                          'task A: [######----]  60% phase 1\n' + \
                          'task A: phase 2\n' + \
                          'task A: [##########] 100% \n' + \
                          'task A: in progress...\n' + \
                          'task A: done\n'

        self.assertEqual(actual_stdout, expected_stdout)
        self.assertEqual(stderr.getvalue(), '')

    def test_label_required(self):
        monitor = ConsoleMonitor()
        with fetch_std_streams():
            with self.assertRaises(ValueError):
                # "ValueError: label must be given"
                monitor.start('', total_work=10)

    def test_child_monitor(self):
        monitor = ConsoleMonitor()
        child_monitor = monitor.child(work=7.5)
        self.assertIsInstance(child_monitor, ChildMonitor)

    def test_cancel(self):
        m = ConsoleMonitor()
        with fetch_std_streams():
            m.start('task A', total_work=10)
            self.assertFalse(m.is_cancelled())
            m.cancel()
            self.assertTrue(m.is_cancelled())


class RecordingMonitorTest(TestCase):
    def test_recording_monitor(self):
        # test the RecordingMonitor used for testing
        m = RecordingMonitor()
        m.start('task A', total_work=10)
        m.progress(work=1)
        m.progress(work=5, msg='phase 1')
        m.progress(msg='phase 2')
        m.progress(work=4)
        m.done()
        self.assertEqual(m.records, [('start', 'task A', 10),
                                     ('progress', 1, None, 10),
                                     ('progress', 5, 'phase 1', 60),
                                     ('progress', None, 'phase 2', None),
                                     ('progress', 4, None, 100),
                                     ('done',)])

    def test_context_manager(self):
        m = RecordingMonitor()
        with m.starting('task A', total_work=10):
            m.progress(work=1)
            m.progress(work=5, msg='phase 1')
            m.progress(msg='phase 2')
            m.progress(work=4)

        self.assertEqual(m.records, [('start', 'task A', 10),
                                     ('progress', 1, None, 10),
                                     ('progress', 5, 'phase 1', 60),
                                     ('progress', None, 'phase 2', None),
                                     ('progress', 4, None, 100),
                                     ('done',)])


class ChildMonitorTest(TestCase):
    def test_child_monitor(self):
        m = RecordingMonitor()
        sub_monitor = m.child(work=10)
        self.assertIsInstance(sub_monitor, ChildMonitor)

        m.start('task A', total_work=10)

        cm1 = m.child(work=1)
        cm1.start('sub-task A.1', 100)
        cm1.progress(work=30)
        cm1.progress(work=20)
        cm1.progress(work=50)
        cm1.done()

        cm2 = m.child(work=5)
        cm2.start('sub-task A.2', 100)
        cm2.progress(work=30)
        cm2.progress(work=20)
        cm2.progress(work=50)
        cm2.done()

        cm3 = m.child(work=4)
        cm3.start('sub-task A.3', 100)
        cm3.progress(work=30)
        cm3.progress(work=20)
        cm3.progress(work=50)
        cm3.done()

        m.done()

        self.assertEqual(m.records, [('start', 'task A', 10),
                                     ('progress', 0.0, 'sub-task A.1', 0),
                                     ('progress', 0.3, None, 3),
                                     ('progress', 0.2, None, 5),
                                     ('progress', 0.5, None, 10),
                                     ('progress', 0.0, 'sub-task A.1', 10),
                                     ('progress', 0.0, 'sub-task A.2', 10),
                                     ('progress', 1.5, None, 25),
                                     ('progress', 1.0, None, 35),
                                     ('progress', 2.5, None, 60),
                                     ('progress', 0.0, 'sub-task A.2', 60),
                                     ('progress', 0.0, 'sub-task A.3', 60),
                                     ('progress', 1.2, None, 72),
                                     ('progress', 0.8, None, 80),
                                     ('progress', 2.0, None, 100),
                                     ('progress', 0.0, 'sub-task A.3', 100),
                                     ('done',)])

    def test_cancel(self):
        m = RecordingMonitor()
        sub_monitor = m.child(work=10)
        self.assertIsInstance(sub_monitor, ChildMonitor)

        m.start('task A', total_work=10)

        cm = m.child(work=1)
        cm.start('sub-task A.1', 100)

        self.assertFalse(cm.is_cancelled())
        self.assertFalse(m.is_cancelled())

        cm.cancel()

        self.assertTrue(cm.is_cancelled())
        self.assertTrue(m.is_cancelled())

    def test_no_label(self):
        m = RecordingMonitor()
        m.start('xxx', 20)
        sm = m.child(work=10)
        sm.start('', total_work=5)
        sm.done()
        sm.start(label=None, total_work=5)
        sm.done()
        m.done()


class RecordingMonitor(Monitor):
    """
    A monitor that buffers progress output as a string so that, e.g.,
    a remote service can pick it up.
    """

    def __init__(self):
        self._records = []
        self._label = None
        self._worked = None
        self._total_work = None
        self._cancelled = None

    @property
    def records(self):
        return self._records

    def start(self, label: str = None, total_work: float = None):
        self._label = label
        self._worked = 0.
        self._total_work = total_work
        self._cancelled = False
        self._records.append(('start', label, total_work))

    def progress(self, work: float = None, msg: str = None):
        self.check_for_cancellation()
        percentage = None
        if work is not None:
            self._worked += work
            percentage = int(100. * self._worked / self._total_work + 0.5)
        self._records.append(('progress', work, msg, percentage))

    def done(self):
        self._records.append(('done',))

    def cancel(self):
        self._records.append(('cancel',))
        self._cancelled = True

    def is_cancelled(self) -> bool:
        return self._cancelled

    def __str__(self):
        return "\n".join([str(r) for r in self._records])



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

The ``esa_climate_toolbox.functions`` package provides functions that represent current
work patterns.

This package is independent of other ``esa_climate_toolbox.*``packages
and can therefore be used stand-alone.

Verification
============

The module's unit-tests are located in
`test/util
<https://github.com/esa-cci/esa-climate-toolbox/tree/main/test/functions>`_ and may
be executed using $ py.test test/util --cov=esa_climate_toolbox/functions``
for extra code coverage information.

Components
==========
"""

__author__ = "Tonio Fincke (Brockmann Consult GmbH)"



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Select variables in a dataset.

Components
==========
"""

import fnmatch

import geopandas as gpd
import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import VarNamesLike


@op(tags=['filter'], version='1.0')
@op_input('ds', data_type=DatasetLike)
@op_input('var', value_set_source='ds', data_type=VarNamesLike)
@op_return(add_history=True)
def select_var(ds: DatasetLike.TYPE,
               var: VarNamesLike.TYPE = None) -> xr.Dataset:
    """
    Filter the dataset, by leaving only the desired variables in it.
    The original dataset information, including original coordinates,
    is preserved.

    :param ds: The dataset or dataframe from which to perform selection.
    :param var: One or more variable names to select and preserve in the
        dataset. All of these are valid 'var_name'
        'var_name1,var_name2,var_name3' ['var_name1', 'var_name2']. One can
        also use wildcards when doing the selection. E.g., choosing 'var_name*'
        for selection will select all variables that start with 'var_name'.
        This can be used to select variables along with their auxiliary
        variables, to select all uncertainty variables, and so on.
    :return: A filtered dataset
    """
    if not var:
        return ds

    ds = DatasetLike.convert(ds)

    var_names = VarNamesLike.convert(var)
    dropped_var_names = list(ds.data_vars.keys())

    for pattern in var_names:
        keep = fnmatch.filter(dropped_var_names, pattern)
        for name in keep:
            dropped_var_names.remove(name)

    return ds.drop_vars(dropped_var_names)


@op(tags=['filter'])
def select_features(df: gpd.GeoDataFrame, var: dict = None) -> gpd.GeoDataFrame:
    """
    Filter the dataframe, by leaving only the desired features in it.
    The original dataframe information, including original features,
    is preserved.

    :param df: The dataframe from which to perform selection.
    :param var: One or more feature names to select and preserve in the
        dataframe.
    :return: A filtered dataframe
    """

    if not var:
        return df

    return df.query(' | '.join(['%s == "%s"' % (key, value)
                                for (key, value) in var.items()]))



import os.path
import sys
from collections import OrderedDict
from unittest import TestCase

import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import op_output
from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.core.op import OpRegistry
from esa_climate_toolbox.core.op import op_return
from esa_climate_toolbox.core.op import new_expression_op
from esa_climate_toolbox.core.op import new_subprocess_op
from esa_climate_toolbox.core.types import FileLike
from esa_climate_toolbox.core.types import VarName
from esa_climate_toolbox.util.misc import object_to_qualified_name
from esa_climate_toolbox.util.monitor import Monitor
from esa_climate_toolbox.util.opmetainf import OpMetaInfo

MONITOR = OpMetaInfo.MONITOR_INPUT_NAME
RETURN = OpMetaInfo.RETURN_OUTPUT_NAME

DIR = os.path.dirname(__file__)
SOILMOISTURE_NC = os.path.join(
    DIR, 'test_data', 'small',
    'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc'
)

MAKE_ENTROPY_EXE = sys.executable + " " + \
                   os.path.join(DIR, 'executables', 'mkentropy.py')
FILTER_DS_EXE = sys.executable + " " + \
                os.path.join(DIR, 'executables', 'filterds.py')


class OpTest(TestCase):

    def setUp(self):
        self.registry = OpRegistry()

    def tearDown(self):
        self.registry = None

    def test_new_executable_op_without_ds(self):
        sub_op = new_subprocess_op(
            OpMetaInfo(
                'make_entropy',
                inputs={
                    'num_steps': {'data_type': int},
                    'period': {'data_type': float},
                },
                outputs={
                    'return': {'data_type': int}
                }
            ),
            MAKE_ENTROPY_EXE + " {num_steps} {period}"
        )
        exit_code = sub_op(num_steps=5, period=0.05)
        self.assertEqual(exit_code, 0)

    def test_new_executable_op_with_ds_file(self):
        sub_op = new_subprocess_op(
            OpMetaInfo(
                'filter_ds',
                inputs={
                    'in_file': {'data_type': FileLike},
                    'out_file': {'data_type': FileLike},
                    'var': {'data_type': VarName},
                },
                outputs={
                    'return': {'data_type': int}
                }),
            FILTER_DS_EXE + " {in_file} {out_file} {var}"
        )
        ofile = os.path.join(DIR, 'test_data', 'filter_ds.nc')
        if os.path.isfile(ofile):
            os.remove(ofile)
        exit_code = sub_op(in_file=SOILMOISTURE_NC, out_file=ofile, var='sm')
        self.assertEqual(exit_code, 0)
        self.assertTrue(os.path.isfile(ofile))
        os.remove(ofile)

    def test_new_executable_op_with_ds_in_mem(self):
        sub_op = new_subprocess_op(
            OpMetaInfo(
                'filter_ds',
                inputs={
                    'ds': {
                        'data_type': xr.Dataset,
                        'write_to': 'in_file'
                    },
                    'var': {'data_type': VarName},
                },
                outputs={
                    'return': {
                        'data_type': xr.Dataset,
                        'read_from': 'out_file'
                    }
                }),
            FILTER_DS_EXE + " {in_file} {out_file} {var}")
        ds = xr.open_dataset(SOILMOISTURE_NC)
        ds_out = sub_op(ds=ds, var='sm')
        self.assertIsNotNone(ds_out)
        self.assertIsNotNone('sm' in ds_out)

    def test_new_expression_op(self):
        expression_op = new_expression_op(
            OpMetaInfo('add_xy',
                       inputs={
                           'x': {'data_type': float},
                           'y': {'data_type': float},
                       },
                       outputs={
                           'return': {'data_type': float}
                       }),
            'x + y'
        )
        z = expression_op(x=1.2, y=2.4)
        self.assertEqual(z, 1.2 + 2.4)

        expression_op = new_expression_op(
            OpMetaInfo(
                'add_xy',
                inputs={'x': {}, 'y': {}}),
            'x * y'
        )
        z = expression_op(x=1.2, y=2.4)
        self.assertEqual(z, 1.2 * 2.4)
        self.assertIn('return', expression_op.op_meta_info.outputs)

    def test_plain_function(self):
        def f(a: float, b, c, u=3, v='A', w=4.9) -> str:
            """Hi, I am f!"""
            return str(a + b + c + u + len(v) + w)

        registry = self.registry
        added_op_reg = registry.add_op(f)
        self.assertIsNotNone(added_op_reg)

        with self.assertRaises(ValueError):
            registry.add_op(f, fail_if_exists=True)

        self.assertIs(registry.add_op(f, fail_if_exists=False), added_op_reg)

        op_reg = registry.get_op(object_to_qualified_name(f))
        self.assertIs(op_reg, added_op_reg)
        self.assertIs(op_reg.wrapped_op, f)
        expected_inputs = OrderedDict()
        expected_inputs['a'] = dict(position=0, data_type=float)
        expected_inputs['b'] = dict(position=1)
        expected_inputs['c'] = dict(position=2)
        expected_inputs['u'] = dict(position=3, default_value=3, data_type=int)
        expected_inputs['v'] = dict(
            position=4, default_value='A', data_type=str
        )
        expected_inputs['w'] = dict(
            position=5, default_value=4.9, data_type=float
        )
        expected_outputs = OrderedDict()
        expected_outputs[RETURN] = dict(data_type=str)
        self._assertMetaInfo(op_reg.op_meta_info,
                             object_to_qualified_name(f),
                             dict(description='Hi, I am f!'),
                             expected_inputs,
                             expected_outputs)

        removed_op_reg = registry.remove_op(f)
        self.assertIs(removed_op_reg, op_reg)
        op_reg = registry.get_op(object_to_qualified_name(f))
        self.assertIsNone(op_reg)

        with self.assertRaises(ValueError):
            registry.remove_op(f, fail_if_not_exists=True)

    def test_decorated_function(self):
        @op(registry=self.registry)
        def f_op(a: float, b, c, u=3, v='A', w=4.9) -> str:
            """Hi, I am f_op!"""
            return str(a + b + c + u + len(v) + w)

        with self.assertRaises(ValueError):
            # must exist
            self.registry.add_op(f_op, fail_if_exists=True)

        op_reg = self.registry.get_op(object_to_qualified_name(f_op))
        expected_inputs = OrderedDict()
        expected_inputs['a'] = dict(position=0, data_type=float)
        expected_inputs['b'] = dict(position=1)
        expected_inputs['c'] = dict(position=2)
        expected_inputs['u'] = dict(
            position=3, default_value=3, data_type=int
        )
        expected_inputs['v'] = dict(
            position=4, default_value='A', data_type=str
        )
        expected_inputs['w'] = dict(
            position=5, default_value=4.9, data_type=float
        )
        expected_outputs = OrderedDict()
        expected_outputs[RETURN] = dict(data_type=str)
        self._assertMetaInfo(op_reg.op_meta_info,
                             object_to_qualified_name(f_op),
                             dict(description='Hi, I am f_op!'),
                             expected_inputs,
                             expected_outputs)

    def test_decorated_function_with_inputs_and_outputs(self):
        @op_input('a', value_range=[0., 1.], registry=self.registry)
        @op_input('v', value_set=['A', 'B', 'C'], registry=self.registry)
        @op_return(registry=self.registry)
        def f_op_inp_ret(a: float, b, c, u=3, v='A', w=4.9) -> str:
            """Hi, I am f_op_inp_ret!"""
            return str(a + b + c + u + len(v) + w)

        with self.assertRaises(ValueError):
            # must exist
            self.registry.add_op(f_op_inp_ret, fail_if_exists=True)

        op_reg = self.registry.get_op(object_to_qualified_name(f_op_inp_ret))
        expected_inputs = OrderedDict()
        expected_inputs['a'] = dict(
            position=0, data_type=float, value_range=[0., 1.]
        )
        expected_inputs['b'] = dict(position=1)
        expected_inputs['c'] = dict(position=2)
        expected_inputs['u'] = dict(position=3, default_value=3, data_type=int)
        expected_inputs['v'] = dict(
            position=4, default_value='A', data_type=str,
            value_set=['A', 'B', 'C']
        )
        expected_inputs['w'] = dict(
            position=5, default_value=4.9, data_type=float
        )
        expected_outputs = OrderedDict()
        expected_outputs[RETURN] = dict(data_type=str)
        self._assertMetaInfo(op_reg.op_meta_info,
                             object_to_qualified_name(f_op_inp_ret),
                             dict(description='Hi, I am f_op_inp_ret!'),
                             expected_inputs,
                             expected_outputs)

    def _assertMetaInfo(self, op_meta_info: OpMetaInfo,
                        expected_name: str,
                        expected_header: dict,
                        expected_input: OrderedDict,
                        expected_output: OrderedDict):
        self.assertIsNotNone(op_meta_info)
        self.assertEqual(op_meta_info.qualified_name, expected_name)
        self.assertEqual(op_meta_info.header, expected_header)
        self.assertEqual(OrderedDict(op_meta_info.inputs), expected_input)
        self.assertEqual(OrderedDict(op_meta_info.outputs), expected_output)

    def test_function_validation(self):
        @op_input('x', registry=self.registry, data_type=float,
                  value_range=[0.1, 0.9], default_value=0.5)
        @op_input('y', registry=self.registry)
        @op_input('a', registry=self.registry, data_type=int,
                  value_set=[1, 4, 5])
        @op_return(registry=self.registry, data_type=float)
        def f(x, y: float or None, a=4):
            return a * x + y if a != 5 else 'foo'

        self.assertIs(f, self.registry.get_op(f))
        self.assertEqual(
            f.op_meta_info.inputs['x'].get('data_type', None), float
        )
        self.assertEqual(
            f.op_meta_info.inputs['x'].get('value_range', None), [0.1, 0.9]
        )
        self.assertEqual(
            f.op_meta_info.inputs['x'].get('default_value', None), 0.5
        )
        self.assertEqual(
            f.op_meta_info.inputs['x'].get('position', None), 0
        )
        self.assertEqual(
            f.op_meta_info.inputs['y'].get('data_type', None), float
        )
        self.assertEqual(
            f.op_meta_info.inputs['y'].get('position', None), 1
        )
        self.assertEqual(
            f.op_meta_info.inputs['a'].get('data_type', None), int
        )
        self.assertEqual(
            f.op_meta_info.inputs['a'].get('value_set', None), [1, 4, 5]
        )
        self.assertEqual(
            f.op_meta_info.inputs['a'].get('default_value', None), 4
        )
        self.assertEqual(
            f.op_meta_info.inputs['a'].get('position', None), 2
        )
        self.assertEqual(
            f.op_meta_info.outputs[RETURN].get('data_type', None), float
        )

        self.assertEqual(f(y=1, x=0.2), 4 * 0.2 + 1)
        self.assertEqual(f(y=3), 4 * 0.5 + 3)
        self.assertEqual(f(0.6, y=3, a=1), 1 * 0.6 + 3.0)

        with self.assertRaises(ValueError) as cm:
            f(y=1, x=8)
        self.assertEqual(
            str(cm.exception),
            "Input 'x' for operation 'test.core.test_op.f' "
            "must be in range [0.1, 0.9]."
        )

        with self.assertRaises(ValueError) as cm:
            f(y=None, x=0.2)
        self.assertEqual(
            str(cm.exception),
            "Input 'y' for operation 'test.core.test_op.f' must be given."
        )

        with self.assertRaises(ValueError) as cm:
            f(y=0.5, x=0.2, a=2)
        self.assertEqual(
            str(cm.exception),
            "Input 'a' for operation 'test.core.test_op.f' "
            "must be one of [1, 4, 5]."
        )

        with self.assertRaises(ValueError) as cm:
            f(x=0, y=3.)
        self.assertEqual(
            str(cm.exception),
            "Input 'x' for operation 'test.core.test_op.f' "
            "must be in range [0.1, 0.9]."
        )

        with self.assertRaises(ValueError) as cm:
            f(x='A', y=3.)
        self.assertEqual(
            str(cm.exception),
            "Input 'x' for operation 'test.core.test_op.f' "
            "must be of type 'float', but got type 'str'."
        )

        with self.assertRaises(ValueError) as cm:
            f(x=0.4)
        self.assertEqual(
            str(cm.exception),
            "Input 'y' for operation 'test.core.test_op.f' must be given."
        )

        with self.assertRaises(ValueError) as cm:
            f(x=0.6, y=0.1, a=2)
        self.assertEqual(
            str(cm.exception),
            "Input 'a' for operation 'test.core.test_op.f' "
            "must be one of [1, 4, 5]."
        )

        with self.assertRaises(ValueError) as cm:
            f(y=3, a=5)
        self.assertEqual(
            str(cm.exception),
            "Output 'return' for operation 'test.core.test_op.f' "
            "must be of type 'float', but got type 'str'."
        )

    def test_function_invocation(self):
        def f(x, a=4):
            return a * x

        op_reg = self.registry.add_op(f)
        result = op_reg(x=2.5)
        self.assertEqual(result, 4 * 2.5)

    def test_function_invocation_with_monitor(self):
        def f(m: Monitor, x, a=4):
            m.start('f', 23)
            return_value = a * x
            m.done()
            return return_value

        op_reg = self.registry.add_op(f)
        monitor = MyMonitor()
        result = op_reg(x=2.5, m=monitor)
        self.assertEqual(result, 4 * 2.5)
        self.assertEqual(monitor.total_work, 23)
        self.assertEqual(monitor.is_done, True)

    def test_history_op(self):
        """
        Test adding operation signature to output history information.
        """
        from esa_climate_toolbox import __version__

        # Test @op_return
        @op(version='0.9', registry=self.registry)
        @op_return(add_history=True, registry=self.registry)
        def history_op(dataset: xr.Dataset, a=1, b='bilinear'):
            res = dataset.copy()
            return res

        ds = xr.Dataset()

        op_reg = self.registry.get_op(object_to_qualified_name(history_op))
        op_meta_info = op_reg.op_meta_info

        # This is a partial stamp, as the way a dict is stringified is not
        # always the same
        stamp = '\nModified with the ESA Climate Toolbox v' \
                + __version__ + ' ' + op_meta_info.qualified_name + ' v' + \
                op_meta_info.header['version'] + \
                ' \nDefault input values: ' + \
                str(op_meta_info.inputs) + '\nProvided input values: '

        ret_ds = op_reg(dataset=ds, a=2, b='trilinear')
        self.assertTrue(stamp in ret_ds.attrs['history'])
        # Check that a passed value is found in the stamp
        self.assertTrue('trilinear' in ret_ds.attrs['history'])

        # Double line-break indicates that this is a subsequent stamp entry
        stamp2 = '\n\nModified with the ESA Climate Toolbox v' + __version__

        ret_ds = op_reg(dataset=ret_ds, a=4, b='quadrilinear')
        self.assertTrue(stamp2 in ret_ds.attrs['history'])
        # Check that a passed value is found in the stamp
        self.assertTrue('quadrilinear' in ret_ds.attrs['history'])
        # Check that a previous passed value is found in the stamp
        self.assertTrue('trilinear' in ret_ds.attrs['history'])

        # Test @op_output
        @op(version='1.9', registry=self.registry)
        @op_output('name1', add_history=True, registry=self.registry)
        @op_output('name2', add_history=False, registry=self.registry)
        @op_output('name3', registry=self.registry)
        def history_named_op(dataset: xr.Dataset, a=1, b='bilinear'):
            ds1 = dataset.copy()
            ds2 = dataset.copy()
            ds3 = dataset.copy()
            return {'name1': ds1, 'name2': ds2, 'name3': ds3}

        ds = xr.Dataset()

        op_reg = self.registry.get_op(
            object_to_qualified_name(history_named_op)
        )
        op_meta_info = op_reg.op_meta_info

        # This is a partial stamp, as the way a dict is stringified is not
        # always the same
        stamp = '\nModified with the ESA Climate Toolbox v' \
                + __version__ + ' ' + op_meta_info.qualified_name + ' v' + \
                op_meta_info.header['version'] + \
                ' \nDefault input values: ' + \
                str(op_meta_info.inputs) + '\nProvided input values: '

        ret = op_reg(dataset=ds, a=2, b='trilinear')
        # Check that the dataset was stamped
        self.assertTrue(stamp in ret['name1'].attrs['history'])
        # Check that a passed value is found in the stamp
        self.assertTrue('trilinear' in ret['name1'].attrs['history'])
        # Check that none of the other two datasets have been stamped
        with self.assertRaises(KeyError):
            ret['name2'].attrs['history']
        with self.assertRaises(KeyError):
            ret['name3'].attrs['history']

        # Double line-break indicates that this is a subsequent stamp entry
        stamp2 = '\n\nModified with the ESA Climate Toolbox v' + __version__

        ret = op_reg(dataset=ret_ds, a=4, b='quadrilinear')
        self.assertTrue(stamp2 in ret['name1'].attrs['history'])
        # Check that a passed value is found in the stamp
        self.assertTrue('quadrilinear' in ret['name1'].attrs['history'])
        # Check that a previous passed value is found in the stamp
        self.assertTrue('trilinear' in ret['name1'].attrs['history'])
        # Other datasets should have the old history, while 'name1' should be
        # updated
        self.assertTrue(ret['name1'].attrs['history']
                        != ret['name2'].attrs['history'])
        self.assertTrue(ret['name1'].attrs['history']
                        != ret['name3'].attrs['history'])
        self.assertTrue(ret['name2'].attrs['history']
                        == ret['name3'].attrs['history'])

        # Test missing version
        @op(registry=self.registry)
        @op_return(add_history=True, registry=self.registry)
        def history_no_version(dataset: xr.Dataset, a=1, b='bilinear'):
            dataset1 = dataset.copy()
            return dataset1

        ds = xr.Dataset()

        op_reg = self.registry.get_op(
            object_to_qualified_name(history_no_version)
        )
        with self.assertRaises(ValueError) as err:
            op_reg(dataset=ds, a=2, b='trilinear')
        self.assertTrue('Could not add history' in str(err.exception))

        # Test not implemented output type stamping
        @op(version='1.1', registry=self.registry)
        @op_return(add_history=True, registry=self.registry)
        def history_wrong_type(ds: xr.Dataset, a=1, b='bilinear'):
            return "Joke's on you"

        ds = xr.Dataset()
        op_reg = self.registry.get_op(
            object_to_qualified_name(history_wrong_type)
        )
        with self.assertRaises(NotImplementedError) as err:
            op_reg(ds=ds, a=2, b='abc')
        self.assertTrue(
            'Adding history information to an' in str(err.exception)
        )


class DefaultOpRegistryTest(TestCase):
    def test_it(self):
        self.assertIsNotNone(OP_REGISTRY)
        self.assertEqual(repr(OP_REGISTRY), 'OP_REGISTRY')


class MyMonitor(Monitor):

    def __init__(self):
        self.total_work = 0
        self.worked = 0
        self.is_done = False

    def start(self, label: str, total_work: float = None):
        self.total_work = total_work

    def progress(self, work: float = None, msg: str = None):
        self.check_for_cancellation()
        self.worked += work

    def done(self):
        self.is_done = True



import cmath
import math
from unittest import TestCase

from esa_climate_toolbox.util.safe import get_safe_globals, safe_eval


class SafeTest(TestCase):
    def test_get_safe_globals(self):
        globals = get_safe_globals()

        self.assertIsNotNone(globals)

        self.assertEqual(globals.get('min'), min)
        self.assertEqual(globals.get('max'), max)
        self.assertEqual(globals.get('cmath'), cmath)
        self.assertEqual(globals.get('math'), math)

        self.assertEqual(globals.get('__builtins__'), None)
        self.assertEqual(globals.get('eval'), None)
        self.assertEqual(globals.get('exec'), None)

    def test_safe_eval_ok(self):
        self.assertEqual(safe_eval('2 + 1'), 3)
        self.assertEqual(safe_eval('x + 1', dict(x=2)), 3)
        self.assertEqual(safe_eval('"Ha%s" % "Ha"'), "HaHa")

    def test_safe_eval_forbidden(self):
        with self.assertRaises(TypeError):
            safe_eval('eval("3+1")')

        with self.assertRaises(TypeError):
            safe_eval('open("test.txt", "w")')



"""
Tests for arithmetic operations
"""

from unittest import TestCase

from datetime import datetime
import numpy as np
import xarray as xr

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.ops import arithmetics
from esa_climate_toolbox.ops import diff
from esa_climate_toolbox.util.misc import object_to_qualified_name


def assert_dataset_equal(expected, actual):
    # this method is functionally equivalent to
    # `assert expected == actual`, but it checks each aspect
    # of equality separately for easier debugging
    assert expected.equals(actual), (expected, actual)


class TestDsArithmetics(TestCase):
    """
    Test dataset arithmetic operations
    """

    def test_nominal(self):
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        actual = arithmetics(dataset, '+2, -2, *3, /3, *4')
        assert_dataset_equal(expected * 4, actual)

        actual = arithmetics(
            dataset, 'exp, log, *10, log10, *2, log2, +1.5'
        )
        assert_dataset_equal(expected * 2.5, actual)

        actual = arithmetics(dataset, 'exp, -1, log1p, +3')
        assert_dataset_equal(expected * 4, actual)

        with self.assertRaises(ValueError) as err:
            arithmetics(dataset, 'not')
        self.assertTrue('not implemented' in str(err.exception))

    def test_registered(self):
        """
        Test the operation when invoked through the OP_REGISTRY
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(arithmetics))
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        actual = reg_op(ds=dataset, ops='+2, -2, *3, /3, *4')
        assert_dataset_equal(expected * 4, actual)


class TestDiff(TestCase):
    """
    Test taking the difference between two datasets
    """

    def test_diff(self):
        # Test nominal
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})
        actual = diff(dataset, dataset * 2)
        assert_dataset_equal(expected * -1, actual)

        # Test variable mismatch
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'third': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        ds1 = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        actual = diff(ds, ds1)
        assert_dataset_equal(expected, actual)
        actual = diff(ds1, ds)
        assert_dataset_equal(expected, actual)

        # Test date range mismatch
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})

        ds1 = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2003, x, 1) for x in range(1, 13)]})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        actual = diff(ds, ds1)
        assert_dataset_equal(actual, expected)

        actual = diff(ds, ds1.drop_vars(['time', ]))
        expected['time'] = [datetime(2000, x, 1) for x in range(1, 13)]
        assert_dataset_equal(actual, expected)

        # Test broadcasting
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        ds1 = xr.Dataset({
            'first': (['lat', 'lon'], np.ones([45, 90])),
            'second': (['lat', 'lon'], np.ones([45, 90])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})
        actual = diff(ds, ds1)
        assert_dataset_equal(expected, actual)

        ds['time'] = [datetime(2000, x, 1) for x in range(1, 4)]
        expected['time'] = [datetime(2000, x, 1) for x in range(1, 4)]
        actual = diff(ds, ds1)
        assert_dataset_equal(expected, actual)

        ds1 = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 1])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 1])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2001, 1, 1)]})
        actual = diff(ds, ds1)
        assert_dataset_equal(expected, actual)

        ds1 = ds1.squeeze('time')
        ds1['time'] = 1
        actual = diff(ds, ds1)
        assert_dataset_equal(expected, actual)

    def test_registered(self):
        """
        Test the operation when invoked from the OP_REGISTRY
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(diff))
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        expected = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 3])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})
        actual = reg_op(ds=dataset, ds2=dataset * 2)
        assert_dataset_equal(expected * -1, actual)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

import logging
import os.path
import sys
from typing import Any, Optional, Sequence, Union

from .defaults import DEFAULT_DATA_PATH
from .defaults import DEFAULT_RES_PATTERN
from .defaults import DEFAULT_VERSION_DATA_PATH
from .defaults import GLOBAL_CONF_FILE
from .defaults import LOCAL_CONF_FILE
from .defaults import LOCATION_FILE
from .defaults import USER_PREFERENCES_FILE
from .defaults import VERSION_CONF_FILE

_CONFIG = None

_LOG = logging.getLogger('ect')


def get_config_path(name: str, default=None) -> str:
    """
    Get the ``str`` value of the configuration parameter *name* which is
    expected to be a path.
    Any tilde character '~' in the value will be expanded
    to the current user's home directory.

    :param name: The name of the configuration parameter.
    :param default: The default value, if *name* is not defined.
    :return: The value
    """
    value = get_config_value(name, default=default)
    return os.path.expanduser(str(value)) if value is not None else None


def get_config_value(name: str, default=None) -> Any:
    """
    Get the value of the configuration parameter *name*.

    :param name: The name of the configuration parameter.
    :param default: The default value, if *name* is not defined.
    :return: The value
    """
    if not name:
        raise ValueError('name argument must be given')
    return get_config().get(name, default)


def get_data_stores_path() -> str:
    """
    Get the default path to where the ESA Climate Toolbox stores local data
    store information and stores data files synchronized with their remote
    versions.

    :return: Effectively reads the value of the configuration parameter
    ``data_stores_path``, if any. Otherwise return the default value
    ``~/.ect/data_stores``.
    """
    return get_config_path('data_stores_path',
                           os.path.join(DEFAULT_DATA_PATH, 'data_stores'))


def get_default_res_pattern() -> str:
    """
    Get the default prefix for names generated for new workspace resources
    originating from opening data sources or executing workflow steps.
    This prefix is used only if no specific prefix is defined for a given
    operation.
    :return: default resource name prefix.
    """
    default_res_pattern = get_config().get('default_res_pattern')
    if default_res_pattern:
        default_res_pattern = default_res_pattern.strip()
    if not default_res_pattern:
        default_res_pattern = DEFAULT_RES_PATTERN
    return default_res_pattern


def get_http_proxy() -> Optional[str]:
    """
    Get a valid 'http_proxy' value. Log a warning if it is invalid.
    :return: URL string or None.
    """
    http_proxy = get_config().get('http_proxy')

    if http_proxy is None:
        return None

    if isinstance(http_proxy, str):
        if http_proxy == '':
            return None
        if http_proxy.startswith('https:') or http_proxy.startswith('http:'):
            return http_proxy

    _LOG.warning('invalid configuration: http_proxy = %r' % http_proxy)

    return http_proxy


def get_config() -> dict:
    """
    Get the global configuration dictionary of the ESA Climate Toolbox.

    :return: A mutable dictionary containing any Python objects.
    """
    global _CONFIG
    if _CONFIG is None:
        _init_config([GLOBAL_CONF_FILE,
                      VERSION_CONF_FILE,
                      LOCAL_CONF_FILE,
                      USER_PREFERENCES_FILE],
                     [DEFAULT_DATA_PATH,
                      DEFAULT_VERSION_DATA_PATH])

    return _CONFIG


def _init_config(
        config_files: Sequence[str], config_dirs,
        template_module: str = 'esa_climate_toolbox.conf.template'
) -> None:
    """
    Initialize the ESA Climate Toolbox configuration.

    :param config_files: A non-empty sequence of configuration files,
        for example ``["~/.ect/conf.py", "~/.ect/<version>/conf.py",
        "./ect-conf.py"]``. Configuration files are read in the given order.
        If the first file does not exist, it will be created and its content
        will be taken from the given *template_module*.
    :param config_dirs: list of directories in which a location file will be
        written.
    :param template_module: Qualified name of a Python module that serves as a
        configuration template file.
        If given, this file will be copied into the parent directory
        of *default_config_file*.
    """
    _write_location_files(config_dirs)
    new_config = _read_config_files(
        config_files, template_module=template_module
    )
    set_config(new_config if new_config is not None else {})


def _read_config_files(
        config_files: Sequence[str],
        template_module: str = 'esa_climate_toolbox.conf.template'
) -> Union[str, None]:
    """
    Initialize the configuration of the ESA Climate Toolbox.

    :param config_files: A non-empty sequence of configuration files,
        for example ``["~/.ect/conf.py", "~/.ect/<version>/conf.py",
        "./ect-conf.py"]``. Configuration files are read in the given order.
        If the first file does not exist, it will be created and its content
        will be taken from the given *template_module*.
    :param template_module: Qualified name of a Python module that serves as a
        configuration template file. If given, this file will be copied into
        the parent directory of *default_config_file*.
    """
    default_config_file = config_files[0]
    if default_config_file and template_module:
        # Always write a 'conf.py.template' so users can rename and
        # use that instead.
        default_config_dir = os.path.dirname(default_config_file)
        config_template_file = os.path.join(default_config_dir,
                                            'conf.py.template')
        try:
            _write_default_config_file(config_template_file, template_module)
        except (IOError, OSError) as error:
            _LOG.warning(
                'failed writing %s: %s' % (config_template_file, str(error))
            )

        if not os.path.exists(default_config_file):
            try:
                _write_default_config_file(
                    default_config_file, template_module
                )
            except (IOError, OSError) as error:
                _LOG.error(
                    'failed writing %s: %s' % (default_config_file, str(error))
                )

    new_config = None
    for config_file in config_files:
        if config_file and os.path.isfile(config_file):
            config = None
            try:
                config = _read_python_config(config_file)
            except Exception as error:
                _LOG.warning(
                    'failed reading %s: %s' % (config_file, str(error))
                )
            if config is not None:
                if new_config is None:
                    new_config = config
                else:
                    new_config.update(config)
    return new_config


def _write_location_files(config_dirs: Sequence[str]):
    """
    Write a location file into the given directories.

    :param config_dirs: Sequence of directories paths.
    """
    new_config = None
    for config_dir in config_dirs:
        if not os.path.isdir(config_dir):
            try:
                os.makedirs(config_dir, exist_ok=True)
            except Exception as error:
                _LOG.warning(
                    'failed creating %s: %s' % (config_dir, str(error))
                )
        location_file = os.path.join(config_dir, LOCATION_FILE)
        try:
            with open(location_file, 'w') as fp:
                fp.write(sys.prefix)
        except Exception as error:
            _LOG.warning('failed writing %s: %s' % (location_file, str(error)))
    return new_config


def _read_python_config(file):
    """
    Reads a configuration *file* which may contain any Python code.

    :param file: Either a configuration file path or a file pointer.
    :return: A dictionary with all the variable assignments made in the configuration file.
    """

    fp = open(file, 'r') if isinstance(file, str) else file
    try:
        config = {}
        code = compile(
            fp.read(), file if isinstance(file, str) else '<NO FILE>', 'exec'
        )
        exec(code, None, config)
        return config
    finally:
        if fp is not file:
            fp.close()


def _write_default_config_file(config_file: str, template_module: str) -> str:
    config_file = os.path.expanduser(config_file)
    config_dir = os.path.dirname(config_file)
    if config_dir and not os.path.exists(config_dir):
        os.makedirs(config_dir, exist_ok=True)

    with open(config_file, 'w', newline='') as fp:
        import pkgutil
        parts = template_module.split('.')
        template_package = '.'.join(parts[:-1])
        template_file = parts[-1] + '.py'
        template_data = pkgutil.get_data(template_package, template_file)
        text = template_data.decode('utf-8')
        fp.write(text)

    return config_file


def set_config(config: dict, update=False) -> None:
    """
    Direct modification of the configuration.

    Should be used for testing only.

    :param config: The new configuration values
    :param update: Whether to update or replace the configuration.
    """
    global _CONFIG
    if _CONFIG is None:
        if update:
            raise RuntimeError('configuration not initialized')
        else:
            _CONFIG = config
    else:
        _CONFIG.update(config)
